#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <richedit.h>
#include "AkelDLL.h"
#include "ConvFunc.h"
#include "StackFunc.h"
#include "StrFunc.h"
#include "Resources\Resource.h"


//Include conversion functions
#define xatoiA
#define xatoiW
#include "ConvFunc.h"

//Include stack functions
#define StackGetIndex
#define StackGetElement
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackMoveAfter
#define StackMoveBefore
#define StackMoveIndex
#define StackDelete
#define StackClear
#define StackSize
#include "StackFunc.h"

//Include string functions
#define StrReplaceA
#define StrReplaceW
#include "StrFunc.h"

//Defines
#define DLLA_HOTKEYS_CHECKHOTKEY   1
#define DLLA_HOTKEYS_GETHOTKEYNAME 2

#define STRID_NAME           1
#define STRID_COMMAND        2
#define STRID_HOTKEY         3
#define STRID_ADD            4
#define STRID_MODIFY         5
#define STRID_DELETE         6
#define STRID_ALLKEYS        7
#define STRID_MOVEUP         8
#define STRID_MOVEDOWN       9
#define STRID_CLOSE          10
#define STRID_AUTOLOAD       11
#define STRID_SYNTAX_ERROR   12
#define STRID_NAME_EXISTS    13
#define STRID_HOTKEY_EXISTS  14

#define AKDLL_MSGHOTKEYEXISTS (WM_USER + 1)
#define AKDLL_UPDATEBUTTON    (WM_USER + 2)

#define LVI_LIST_NAME     0
#define LVI_LIST_COMMAND  1
#define LVI_LIST_HOTKEY   2

#define BUFFER_SIZE       1024

#define OF_HOTKEYTEXT     0x1
#define OF_WINDOWRECT     0x2

#define ACT_COMMAND       1
#define ACT_CALL          2
#define ACT_EXEC          3
#define ACT_FONT          4
#define ACT_RECODE        5
#define ACT_INSERT        6

#define PARAM_CHAR        1
#define PARAM_INT         2

#define FS_NONE            0
#define FS_FONTNORMAL      1
#define FS_FONTITALIC      2
#define FS_FONTBOLD        3
#define FS_FONTBOLDITALIC  4

//DIALOGRESIZE
#define DRS_SIZE  0x1 //Resize control
#define DRS_MOVE  0x2 //Move control
#define DRS_X     0x4 //X value
#define DRS_Y     0x8 //Y value

#ifndef LVM_SETEXTENDEDLISTVIEWSTYLE
  #define LVM_SETEXTENDEDLISTVIEWSTYLE (LVM_FIRST + 54)
#endif
#ifndef LVS_EX_GRIDLINES
  #define LVS_EX_GRIDLINES 0x00000001
#endif
#ifndef LVS_EX_FULLROWSELECT
  #define LVS_EX_FULLROWSELECT 0x00000020
#endif

#define TXT_DEFAULT_HOTKEYS \
L"\"Case Loop\" Command(4180) Hotkey(0)\r\
\"Notepad\" Exec(\"notepad.exe\") Hotkey(0)"

typedef struct _HOTKEYITEMA {
  struct _HOTKEYITEMA *next;
  struct _HOTKEYITEMA *prev;
  char szHotkeyName[MAX_PATH];
  char szHotkeyCommand[MAX_PATH];
  DWORD dwHotkey;
  BOOL bSetOnStart;
  DWORD dwAction;
  HSTACK hParamStack;
} HOTKEYITEMA;

typedef struct _HOTKEYITEMW {
  struct _HOTKEYITEMW *next;
  struct _HOTKEYITEMW *prev;
  wchar_t wszHotkeyName[MAX_PATH];
  wchar_t wszHotkeyCommand[MAX_PATH];
  DWORD dwHotkey;
  BOOL bSetOnStart;
  DWORD dwAction;
  HSTACK hParamStack;
} HOTKEYITEMW;

typedef struct _PARAMETER {
  struct _PARAMETER *next;
  struct _PARAMETER *prev;
  DWORD dwType;
  LPARAM lParam;
  unsigned char *pExpanded;
} PARAMETER;

typedef struct _PLUGINCALL {
  unsigned char *lpStruct;
  BOOL bOnStart;
  BOOL bSetOnStart;
} PLUGINCALL;

typedef struct {
  int nCmd;        //Command identifier
  char *pCmdName;  //Command name
} CMDLIST;

typedef struct {
  HWND *lpWnd;   //Control window
  DWORD dwType;  //See DRS_* defines
} DIALOGRESIZE;

//Functions prototypes
BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK AllKeysDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK AllKeysDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
void FillHotkeyListA(HWND hWnd, HSTACK *hStack);
void FillHotkeyListW(HWND hWnd, HSTACK *hStack);
int SetHotkeyListItemA(HWND hWnd, int nIndex, HOTKEYITEMA *hiElement, BOOL bAdd);
int SetHotkeyListItemW(HWND hWnd, int nIndex, HOTKEYITEMW *hiElement, BOOL bAdd);
void SelHotkeyListItem(HWND hWnd, int nIndex);
int CALLBACK CompareFuncA(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);
int CALLBACK CompareFuncW(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void CreateHotkeyStackA(HSTACK *hStack, char *pText);
void CreateHotkeyStackW(HSTACK *hStack, wchar_t *wpText);
void CALLBACK CallPluginA(void *lpParameter);
void CALLBACK CallPluginW(void *lpParameter);
BOOL CallHotkeyA(HSTACK *hStack, WORD wHotkey);
BOOL CallHotkeyW(HSTACK *hStack, WORD wHotkey);
HOTKEYITEMA* StackInsertHotkeyA(HSTACK *hStack);
HOTKEYITEMW* StackInsertHotkeyW(HSTACK *hStack);
int StackGetItemIndex(HSTACK *hStack, void *lpElement);
HOTKEYITEMA* StackGetItemByIndexA(HSTACK *hStack, int nIndex);
HOTKEYITEMW* StackGetItemByIndexW(HSTACK *hStack, int nIndex);
HOTKEYITEMA* StackGetItemByHotkeyA(HSTACK *hStack, DWORD dwHotkey);
HOTKEYITEMW* StackGetItemByHotkeyW(HSTACK *hStack, DWORD dwHotkey);
HOTKEYITEMA* StackGetItemByNameA(HSTACK *hStack, char *pName);
HOTKEYITEMW* StackGetItemByNameW(HSTACK *hStack, wchar_t *wpName);
void StackMoveItem(HSTACK *hStack, int nOldIndex, int nNewIndex);
void StackDeleteItemByIndexA(HSTACK *hStack, int nIndex);
void StackDeleteItemByIndexW(HSTACK *hStack, int nIndex);
void StackFreeHotkeyA(HSTACK *hStack);
void StackFreeHotkeyW(HSTACK *hStack);
void GetActionParametersA(HSTACK *hParamStack, char *pText, char **ppText);
void GetActionParametersW(HSTACK *hParamStack, wchar_t *wpText, wchar_t **wppText);
void SetParametersExpCharA(HSTACK *hParamStack, char *pFile, char *pExeDir);
void SetParametersExpCharW(HSTACK *hParamStack, wchar_t *wpFile, wchar_t *wpExeDir);
int CreateParametersStruct(HSTACK *hParamStack, unsigned char **lppStruct);
int GetParameterInt(HSTACK *hParamStack, int nIndex);
unsigned char* GetParameterChar(HSTACK *hParamStack, int nIndex);
unsigned char* GetParameterExpChar(HSTACK *hParamStack, int nIndex);
void FreeActionParameters(HSTACK *hParamStack);
int NextStringA(char *pText, char *szStr, int nStrLen, char **ppText, int *nMinus);
int NextStringW(wchar_t *wpText, wchar_t *wszStr, int nStrLen, wchar_t **wppText, int *nMinus);
int GetActionNameA(char *pText, char *szStr, int nStrLen, char **ppText);
int GetActionNameW(wchar_t *wpText, wchar_t *wszStr, int nStrLen, wchar_t **wppText);
int GetFileDirA(char *pFile, char *szFileDir, int nFileDirLen);
int GetFileDirW(wchar_t *wpFile, wchar_t *wszFileDir, int nFileDirLen);
void GetHotkeyStringA(WORD wHotkey, char *szString);
void GetHotkeyStringW(WORD wHotkey, wchar_t *wszString);
int GetControlTextA(HWND hWnd, char *szText, int nTextMax);
int GetControlTextW(HWND hWnd, wchar_t *wszText, int nTextMax);
BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc);
BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA(DWORD dwFlags);
void SaveOptionsW(DWORD dwFlags);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceEXE;
HINSTANCE hInstanceDLL;
HSTACK *hPluginsStack;
HWND hMainWnd;
HWND hMdiClient;
HMENU hMainMenu;
HMENU hMenuRecentFiles;
HMENU hPopupEdit;
HICON hMainIcon;
HACCEL hGlobalAccel;
HACCEL hMainAccel;
BOOL bOldWindows;
BOOL bMDI;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
char szExeDir[MAX_PATH];
wchar_t wszExeDir[MAX_PATH];
char szBuffer[BUFFER_SIZE];
wchar_t wszBuffer[BUFFER_SIZE];
char *szHotkeyText=NULL;
wchar_t *wszHotkeyText=NULL;
HSTACK hHotkeysStack={0};
RECT rcMainInitDialog={0};
RECT rcMainCurrentDialog={0};
RECT rcAllKeysInitDialog={0};
RECT rcAllKeysCurrentDialog={0};
BOOL bMainDialogRectSave=FALSE;
int nColumnWidth1=145;
int nColumnWidth2=276;
int nColumnWidth3=100;
WNDPROCDATA *NewMainProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="Hotkeys";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (pd->lParam)
  {
    if (bInitMain)
    {
      if (IsExtCallParamValid(pd->lParam, 1))
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_HOTKEYS_CHECKHOTKEY)
        {
          DWORD dwHotkey=0;
          BOOL *lpbExist=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            dwHotkey=GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            lpbExist=(BOOL *)GetExtCallParam(pd->lParam, 3);

          if (lpbExist)
          {
            if (bOldWindows)
            {
              if (StackGetItemByHotkeyA(&hHotkeysStack, dwHotkey))
                *lpbExist=TRUE;
              else
                *lpbExist=FALSE;
            }
            else
            {
              if (StackGetItemByHotkeyW(&hHotkeysStack, dwHotkey))
                *lpbExist=TRUE;
              else
                *lpbExist=FALSE;
            }
          }
        }
        else if (GetExtCallParam(pd->lParam, 1) == DLLA_HOTKEYS_GETHOTKEYNAME)
        {
          DWORD dwHotkey=0;
          unsigned char *szName=NULL;
          int nNameMax=0;

          if (IsExtCallParamValid(pd->lParam, 2))
            dwHotkey=GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            szName=(unsigned char *)GetExtCallParam(pd->lParam, 3);
          if (IsExtCallParamValid(pd->lParam, 4))
            nNameMax=GetExtCallParam(pd->lParam, 4);

          if (szName && nNameMax)
          {
            if (bOldWindows)
            {
              HOTKEYITEMA *lpElement;
              char *pName=(char *)szName;

              if (lpElement=StackGetItemByHotkeyA(&hHotkeysStack, dwHotkey))
                lstrcpynA(pName, lpElement->szHotkeyName, nNameMax);
              else
                pName[0]='\0';
            }
            else
            {
              HOTKEYITEMW *lpElement;
              wchar_t *pName=(wchar_t *)szName;

              if (lpElement=StackGetItemByHotkeyW(&hHotkeysStack, dwHotkey))
                lstrcpynW(pName, lpElement->wszHotkeyName, nNameMax);
              else
                pName[0]='\0';
            }
          }
        }
      }
    }
    else return;
  }
  else
  {
    if (!bInitCommon) InitCommon(pd);
    if (!bInitMain) InitMain();

    if (!pd->bOnStart)
    {
      if (bOldWindows)
        DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_MAIN), hMainWnd, (DLGPROC)MainDlgProcA);
      else
        DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_MAIN), hMainWnd, (DLGPROC)MainDlgProcW);
    }
  }

  //Stay in memory, and show as active
  pd->nUnload=UD_NONUNLOAD_ACTIVE;
}

BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndName;
  static HWND hWndCommand;
  static HWND hWndHotkeyLabel;
  static HWND hWndHotkey;
  static HWND hWndButtonGroup;
  static HWND hWndAddModifyHotkey;
  static HWND hWndMoveUpHotkey;
  static HWND hWndMoveDownHotkey;
  static HWND hWndDeleteHotkey;
  static HWND hWndList;
  static HWND hWndAllKeys;
  static HWND hWndClose;
  static int nSelItem=-1;
  static BOOL bListChanged=FALSE;
  static DIALOGRESIZE drs[]={{&hWndList,            DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndCommand,         DRS_SIZE|DRS_X},
                             {&hWndHotkeyLabel,     DRS_MOVE|DRS_X},
                             {&hWndHotkey,          DRS_MOVE|DRS_X},
                             {&hWndButtonGroup,     DRS_MOVE|DRS_X},
                             {&hWndAddModifyHotkey, DRS_MOVE|DRS_X},
                             {&hWndMoveUpHotkey,    DRS_MOVE|DRS_X},
                             {&hWndMoveDownHotkey,  DRS_MOVE|DRS_X},
                             {&hWndDeleteHotkey,    DRS_MOVE|DRS_X},
                             {&hWndAllKeys,         DRS_MOVE|DRS_X},
                             {&hWndClose,           DRS_MOVE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNA lvc;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndList=GetDlgItem(hDlg, IDC_LIST);
    hWndName=GetDlgItem(hDlg, IDC_NAME);
    hWndCommand=GetDlgItem(hDlg, IDC_COMMAND);
    hWndHotkeyLabel=GetDlgItem(hDlg, IDC_HOTKEY_LABEL);
    hWndHotkey=GetDlgItem(hDlg, IDC_HOTKEY);
    hWndButtonGroup=GetDlgItem(hDlg, IDC_BUTTONGROUP);
    hWndAddModifyHotkey=GetDlgItem(hDlg, IDC_ADDMODIFYHOTKEY);
    hWndMoveUpHotkey=GetDlgItem(hDlg, IDC_MOVEUPHOTKEY);
    hWndMoveDownHotkey=GetDlgItem(hDlg, IDC_MOVEDOWNHOTKEY);
    hWndDeleteHotkey=GetDlgItem(hDlg, IDC_DELETEHOTKEY);
    hWndAllKeys=GetDlgItem(hDlg, IDC_ALLKEYS);
    hWndClose=GetDlgItem(hDlg, IDC_CLOSE);

    SetDlgItemTextA(hDlg, IDC_NAME_LABEL, GetLangStringA(wLangSystem, STRID_NAME));
    SetDlgItemTextA(hDlg, IDC_COMMAND_LABEL, GetLangStringA(wLangSystem, STRID_COMMAND));
    SetDlgItemTextA(hDlg, IDC_HOTKEY_LABEL, GetLangStringA(wLangSystem, STRID_HOTKEY));
    SetDlgItemTextA(hDlg, IDC_ADDMODIFYHOTKEY, GetLangStringA(wLangSystem, STRID_ADD));
    SetDlgItemTextA(hDlg, IDC_DELETEHOTKEY, GetLangStringA(wLangSystem, STRID_DELETE));
    SetDlgItemTextA(hDlg, IDC_MOVEUPHOTKEY, GetLangStringA(wLangSystem, STRID_MOVEUP));
    SetDlgItemTextA(hDlg, IDC_MOVEDOWNHOTKEY, GetLangStringA(wLangSystem, STRID_MOVEDOWN));
    SetDlgItemTextA(hDlg, IDC_ALLKEYS, GetLangStringA(wLangSystem, STRID_ALLKEYS));
    SetDlgItemTextA(hDlg, IDC_CLOSE, GetLangStringA(wLangSystem, STRID_CLOSE));

    SendMessage(hWndName, EM_LIMITTEXT, MAX_PATH, 0);
    SendMessage(hWndCommand, EM_LIMITTEXT, MAX_PATH, 0);
    EnableWindow(hWndAddModifyHotkey, FALSE);
    EnableWindow(hWndMoveUpHotkey, FALSE);
    EnableWindow(hWndMoveDownHotkey, FALSE);
    EnableWindow(hWndDeleteHotkey, FALSE);
    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES);

    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndHotkey, 0);

    //Columns
    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_NAME);
    lvc.cx=nColumnWidth1;
    lvc.iSubItem=LVI_LIST_NAME;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_LIST_NAME, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_COMMAND);
    lvc.cx=nColumnWidth2;
    lvc.iSubItem=LVI_LIST_COMMAND;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_LIST_COMMAND, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_HOTKEY);
    lvc.cx=nColumnWidth3;
    lvc.iSubItem=LVI_LIST_HOTKEY;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_LIST_HOTKEY, (LPARAM)&lvc);

    SendMessageA(hWndCommand, CB_ADDSTRING, 0, (LPARAM)"Command(0)");
    SendMessageA(hWndCommand, CB_ADDSTRING, 0, (LPARAM)"Exec(`Program.exe`)");
    SendMessageA(hWndCommand, CB_ADDSTRING, 0, (LPARAM)"Call(\"Plugin::Function\")");
    SendMessageA(hWndCommand, CB_ADDSTRING, 0, (LPARAM)"+Call(\"Plugin::Function\")");
    SendMessageA(hWndCommand, CB_ADDSTRING, 0, (LPARAM)"Font(\"FaceName\", 0, 10)");
    SendMessageA(hWndCommand, CB_ADDSTRING, 0, (LPARAM)"Recode(0, 0)");
    SendMessageA(hWndCommand, CB_ADDSTRING, 0, (LPARAM)"Insert(\"text\")");

    FillHotkeyListA(hWndList, &hHotkeysStack);
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_LIST)
    {
      HOTKEYITEMA *lpElement;

      if (((NMLISTVIEW *)lParam)->hdr.code == LVN_ITEMCHANGED)
      {
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_SELECTED)
        {
          nSelItem=((NMLISTVIEW *)lParam)->iItem;

          if (lpElement=StackGetItemByIndexA(&hHotkeysStack, nSelItem + 1))
          {
            if (nSelItem == 0)
              EnableWindow(hWndMoveUpHotkey, FALSE);
            else
              EnableWindow(hWndMoveUpHotkey, TRUE);
            if (SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0) == nSelItem + 1)
              EnableWindow(hWndMoveDownHotkey, FALSE);
            else
              EnableWindow(hWndMoveDownHotkey, TRUE);
            EnableWindow(hWndDeleteHotkey, TRUE);

            SetWindowTextA(hWndName, lpElement->szHotkeyName);
            SetWindowTextA(hWndCommand, lpElement->szHotkeyCommand);
            SendMessage(hWndHotkey, HKM_SETHOTKEY, lpElement->dwHotkey, 0);
            PostMessage(hDlg, AKDLL_UPDATEBUTTON, 0, 0);
          }
        }
        if (((NMLISTVIEW *)lParam)->uOldState & LVIS_SELECTED)
        {
          nSelItem=-1;

          EnableWindow(hWndMoveUpHotkey, FALSE);
          EnableWindow(hWndMoveDownHotkey, FALSE);
          EnableWindow(hWndDeleteHotkey, FALSE);
          PostMessage(hDlg, AKDLL_UPDATEBUTTON, 0, 0);
        }
      }
      else if ((int)((NMHDR *)lParam)->code == NM_DBLCLK)
      {
        if (nSelItem >= 0)
          SendMessage(hWndList, LVM_EDITLABEL, nSelItem, 0);
      }
      else if (((NMHDR *)lParam)->code == LVN_ENDLABELEDITA)
      {
        NMLVDISPINFOA *nmlvdi=(NMLVDISPINFOA *)lParam;

        if (nmlvdi->item.pszText && nmlvdi->item.pszText[0])
        {
          if (!StackGetItemByNameA(&hHotkeysStack, nmlvdi->item.pszText))
          {
            if (lpElement=StackGetItemByIndexA(&hHotkeysStack, nSelItem + 1))
            {
              lstrcpynA(lpElement->szHotkeyName, nmlvdi->item.pszText, MAX_PATH);
              SetWindowLongA(hDlg, DWL_MSGRESULT, 1);
            }
          }
          else PostMessage(hDlg, AKDLL_MSGHOTKEYEXISTS, 0, 0);
        }
        return TRUE;
      }
    }
  }
  else if (uMsg == AKDLL_MSGHOTKEYEXISTS)
  {
    MessageBoxA(hDlg, GetLangStringA(wLangSystem, STRID_NAME_EXISTS), "Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);
  }
  else if (uMsg == AKDLL_UPDATEBUTTON)
  {
    char szHotkeyName[MAX_PATH];
    char szHotkeyCommand[MAX_PATH];
    HOTKEYITEMA *lpElement;
    HWND hWndFocus;
    WORD wHotkey;
    BOOL bEnable=FALSE;

    hWndFocus=GetFocus();
    SetDlgItemTextA(hDlg, IDC_ADDMODIFYHOTKEY, GetLangStringA(wLangSystem, STRID_ADD));

    if (GetControlTextA(hWndName, szHotkeyName, MAX_PATH))
    {
      GetControlTextA(hWndCommand, szHotkeyCommand, MAX_PATH);
      wHotkey=(WORD)SendMessage(hWndHotkey, HKM_GETHOTKEY, 0, 0);
      bEnable=TRUE;

      if (lpElement=StackGetItemByNameA(&hHotkeysStack, szHotkeyName))
      {
        SetDlgItemTextA(hDlg, IDC_ADDMODIFYHOTKEY, GetLangStringA(wLangSystem, STRID_MODIFY));
        if (!lstrcmpA(szHotkeyCommand, lpElement->szHotkeyCommand) && lpElement->dwHotkey == wHotkey)
          bEnable=FALSE;
      }
    }

    if (!bEnable)
    {
      if (hWndFocus == hWndAddModifyHotkey)
        SetFocus(hWndList);
    }
    EnableWindow(hWndAddModifyHotkey, bEnable);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_NAME ||
        LOWORD(wParam) == IDC_COMMAND ||
        LOWORD(wParam) == IDC_HOTKEY)
    {
      PostMessage(hDlg, AKDLL_UPDATEBUTTON, 0, 0);
    }
    else if (LOWORD(wParam) == IDC_ADDMODIFYHOTKEY)
    {
      PLUGINFUNCTIONA *pfElement=NULL;
      HOTKEYITEMA *hiNewElement=NULL;
      HOTKEYITEMA *hiOldElement=NULL;
      HOTKEYITEMA *hiElement=NULL;
      char szHotkeyName[MAX_PATH];
      char szHotkeyCommand[MAX_PATH];
      char *pText=szHotkeyCommand;
      char *pHotkeyCommand;
      DWORD dwAction;
      WORD wHotkey;
      int nNewIndex;
      int nOldIndex=-1;
      int nPlus;
      int a;
      int b;
      BOOL bAddHotkey=(LOWORD(wParam) == IDC_ADDMODIFYHOTKEY);

      GetControlTextA(hWndName, szHotkeyName, MAX_PATH);
      GetControlTextA(hWndCommand, szHotkeyCommand, MAX_PATH);
      wHotkey=(WORD)SendMessage(hWndHotkey, HKM_GETHOTKEY, 0, 0);
      if (hiOldElement=StackGetItemByNameA(&hHotkeysStack, szHotkeyName))
      {
        bAddHotkey=FALSE;
        nOldIndex=StackGetItemIndex(&hHotkeysStack, hiOldElement) - 1;
      }

      //Remove quotes
      for (a=0, b=0; szHotkeyName[b];)
      {
        if (szHotkeyName[b] == '\"')
          ++b;
        else
          szHotkeyName[a++]=szHotkeyName[b++];
      }
      szHotkeyName[a]='\0';

      if (wHotkey)
      {
        //Is hotkey already assigned via AKD_DLLADD
        if (pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, wHotkey))
        {
          wsprintfA(szBuffer, GetLangStringA(wLangSystem, STRID_HOTKEY_EXISTS), pfElement->szFunction);
          MessageBoxA(hDlg, szBuffer, "Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);
          SetFocus(hWndHotkey);
          return FALSE;
        }

        //Is hotkey already assigned via HotKeys plugin
        if (hiElement=StackGetItemByHotkeyA(&hHotkeysStack, wHotkey))
        {
          if (bAddHotkey || hiOldElement != hiElement)
          {
            wsprintfA(szBuffer, GetLangStringA(wLangSystem, STRID_HOTKEY_EXISTS), hiElement->szHotkeyName);
            MessageBoxA(hDlg, szBuffer, "Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);
            SetFocus(hWndHotkey);
            return FALSE;
          }
        }
      }

      //AutoLoad
      while (*pText == ' ') ++pText;
      pHotkeyCommand=pText;

      if (*pText == '+')
      {
        ++pText;
        nPlus=1;
      }
      else nPlus=0;

      //Command
      dwAction=0;
      GetActionNameA(pText, szBuffer, BUFFER_SIZE, &pText);

      if (!lstrcmpiA(szBuffer, "Command"))
        dwAction=ACT_COMMAND;
      else if (!lstrcmpiA(szBuffer, "Call"))
        dwAction=ACT_CALL;
      else if (!lstrcmpiA(szBuffer, "Exec"))
        dwAction=ACT_EXEC;
      else if (!lstrcmpiA(szBuffer, "Font"))
        dwAction=ACT_FONT;
      else if (!lstrcmpiA(szBuffer, "Recode"))
        dwAction=ACT_RECODE;
      else if (!lstrcmpiA(szBuffer, "Insert"))
        dwAction=ACT_INSERT;

      if (dwAction)
      {
        if (bAddHotkey)
        {
          if (hiNewElement=StackInsertHotkeyA(&hHotkeysStack))
          {
            lstrcpynA(hiNewElement->szHotkeyName, szHotkeyName, MAX_PATH);
            hiNewElement->szHotkeyCommand[0]='\0';
            hiNewElement->dwHotkey=wHotkey;
            hiNewElement->bSetOnStart=nPlus;
            hiNewElement->dwAction=dwAction;
            GetActionParametersA(&hiNewElement->hParamStack, pText, &pText);

            //Command string
            lstrcpynA(hiNewElement->szHotkeyCommand, pHotkeyCommand, min(pText - pHotkeyCommand + 1, MAX_PATH));
          }
        }
        else
        {
          hiNewElement=hiOldElement;

          if (hiNewElement)
          {
            lstrcpynA(hiNewElement->szHotkeyName, szHotkeyName, MAX_PATH);
            hiNewElement->szHotkeyCommand[0]='\0';
            hiNewElement->dwHotkey=wHotkey;
            hiNewElement->bSetOnStart=nPlus;
            hiNewElement->dwAction=dwAction;
            FreeActionParameters(&hiNewElement->hParamStack);
            GetActionParametersA(&hiNewElement->hParamStack, pText, &pText);

            //Command string
            lstrcpynA(hiNewElement->szHotkeyCommand, pHotkeyCommand, min(pText - pHotkeyCommand + 1, MAX_PATH));
          }
        }
      }
      else MessageBoxA(hDlg, GetLangStringA(wLangSystem, STRID_SYNTAX_ERROR), "Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);

      if ((nNewIndex=SetHotkeyListItemA(hWndList, nOldIndex, hiNewElement, bAddHotkey)) >= 0)
      {
        SelHotkeyListItem(hWndList, nNewIndex);
        bListChanged=TRUE;
      }
      PostMessage(hDlg, AKDLL_UPDATEBUTTON, 0, 0);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_MOVEUPHOTKEY ||
             LOWORD(wParam) == IDC_MOVEDOWNHOTKEY)
    {
      HOTKEYITEMA *lpElement;
      int nOldIndex=nSelItem;
      int nNewIndex;

      if (nOldIndex != -1)
      {
        if (lpElement=StackGetItemByIndexA(&hHotkeysStack, nOldIndex + 1))
        {
          //Delete
          SendMessage(hWndList, LVM_DELETEITEM, nOldIndex, 0);

          //Add
          if (LOWORD(wParam) == IDC_MOVEUPHOTKEY)
            nNewIndex=max(nOldIndex - 1, 0);
          else
            nNewIndex=nOldIndex + 1;
          nNewIndex=SetHotkeyListItemA(hWndList, nNewIndex, lpElement, TRUE);

          //Move item in stack
          StackMoveItem(&hHotkeysStack, nOldIndex + 1, nNewIndex + 1);

          //Select
          SetFocus(hWndList);
          SelHotkeyListItem(hWndList, nNewIndex);

          bListChanged=TRUE;
        }
      }
    }
    else if (LOWORD(wParam) == IDC_DELETEHOTKEY)
    {
      int nIndex=nSelItem;

      if (nSelItem != -1)
      {
        //Delete
        StackDeleteItemByIndexA(&hHotkeysStack, nIndex + 1);
        SendMessage(hWndList, LVM_DELETEITEM, nIndex, 0);
        nIndex=min(nIndex, SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0) - 1);

        //Select
        SetFocus(hWndList);
        SelHotkeyListItem(hWndList, nIndex);
        bListChanged=TRUE;
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      if (GetFocus() == hWndList)
        SendMessage(hWndList, LVM_EDITLABEL, nSelItem, 0);
    }
    else if (LOWORD(wParam) == IDC_ALLKEYS)
    {
      DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_ALLKEYS), hDlg, (DLGPROC)AllKeysDlgProcA);
    }
    else if (LOWORD(wParam) == IDC_CLOSE ||
             LOWORD(wParam) == IDCANCEL)
    {
      int nWidth;

      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_LIST_NAME, 0);
      if (nWidth != nColumnWidth1)
      {
        nColumnWidth1=nWidth;
        bMainDialogRectSave=TRUE;
      }
      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_LIST_COMMAND, 0);
      if (nWidth != nColumnWidth2)
      {
        nColumnWidth2=nWidth;
        bMainDialogRectSave=TRUE;
      }
      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_LIST_HOTKEY, 0);
      if (nWidth != nColumnWidth3)
      {
        nColumnWidth3=nWidth;
        bMainDialogRectSave=TRUE;
      }

      if (bListChanged)
      {
        HOTKEYITEMA *lpElement;
        DWORD dwSize;

        for (lpElement=(HOTKEYITEMA *)hHotkeysStack.first, dwSize=0; lpElement; lpElement=lpElement->next)
        {
          dwSize+=wsprintfA(szBuffer, "\"%s\" %s Hotkey(%d)\r", lpElement->szHotkeyName, lpElement->szHotkeyCommand, lpElement->dwHotkey);
        }
        if (szHotkeyText=(char *)GlobalAlloc(GPTR, dwSize + 1))
        {
          for (lpElement=(HOTKEYITEMA *)hHotkeysStack.first, dwSize=0; lpElement; lpElement=lpElement->next)
          {
            dwSize+=wsprintfA(szHotkeyText + dwSize, "\"%s\" %s Hotkey(%d)\r", lpElement->szHotkeyName, lpElement->szHotkeyCommand, lpElement->dwHotkey);
          }
          SaveOptionsA(OF_HOTKEYTEXT);
          GlobalFree((HGLOBAL)szHotkeyText);
        }
      }
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndName;
  static HWND hWndCommand;
  static HWND hWndHotkeyLabel;
  static HWND hWndHotkey;
  static HWND hWndButtonGroup;
  static HWND hWndAddModifyHotkey;
  static HWND hWndMoveUpHotkey;
  static HWND hWndMoveDownHotkey;
  static HWND hWndDeleteHotkey;
  static HWND hWndList;
  static HWND hWndAllKeys;
  static HWND hWndClose;
  static int nSelItem=-1;
  static BOOL bListChanged=FALSE;
  static DIALOGRESIZE drs[]={{&hWndList,            DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndCommand,         DRS_SIZE|DRS_X},
                             {&hWndHotkeyLabel,     DRS_MOVE|DRS_X},
                             {&hWndHotkey,          DRS_MOVE|DRS_X},
                             {&hWndButtonGroup,     DRS_MOVE|DRS_X},
                             {&hWndAddModifyHotkey, DRS_MOVE|DRS_X},
                             {&hWndMoveUpHotkey,    DRS_MOVE|DRS_X},
                             {&hWndMoveDownHotkey,  DRS_MOVE|DRS_X},
                             {&hWndDeleteHotkey,    DRS_MOVE|DRS_X},
                             {&hWndAllKeys,         DRS_MOVE|DRS_X},
                             {&hWndClose,           DRS_MOVE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNW lvc;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndList=GetDlgItem(hDlg, IDC_LIST);
    hWndName=GetDlgItem(hDlg, IDC_NAME);
    hWndCommand=GetDlgItem(hDlg, IDC_COMMAND);
    hWndHotkeyLabel=GetDlgItem(hDlg, IDC_HOTKEY_LABEL);
    hWndHotkey=GetDlgItem(hDlg, IDC_HOTKEY);
    hWndButtonGroup=GetDlgItem(hDlg, IDC_BUTTONGROUP);
    hWndAddModifyHotkey=GetDlgItem(hDlg, IDC_ADDMODIFYHOTKEY);
    hWndMoveUpHotkey=GetDlgItem(hDlg, IDC_MOVEUPHOTKEY);
    hWndMoveDownHotkey=GetDlgItem(hDlg, IDC_MOVEDOWNHOTKEY);
    hWndDeleteHotkey=GetDlgItem(hDlg, IDC_DELETEHOTKEY);
    hWndAllKeys=GetDlgItem(hDlg, IDC_ALLKEYS);
    hWndClose=GetDlgItem(hDlg, IDC_CLOSE);

    SetDlgItemTextW(hDlg, IDC_NAME_LABEL, GetLangStringW(wLangSystem, STRID_NAME));
    SetDlgItemTextW(hDlg, IDC_COMMAND_LABEL, GetLangStringW(wLangSystem, STRID_COMMAND));
    SetDlgItemTextW(hDlg, IDC_HOTKEY_LABEL, GetLangStringW(wLangSystem, STRID_HOTKEY));
    SetDlgItemTextW(hDlg, IDC_ADDMODIFYHOTKEY, GetLangStringW(wLangSystem, STRID_ADD));
    SetDlgItemTextW(hDlg, IDC_DELETEHOTKEY, GetLangStringW(wLangSystem, STRID_DELETE));
    SetDlgItemTextW(hDlg, IDC_MOVEUPHOTKEY, GetLangStringW(wLangSystem, STRID_MOVEUP));
    SetDlgItemTextW(hDlg, IDC_MOVEDOWNHOTKEY, GetLangStringW(wLangSystem, STRID_MOVEDOWN));
    SetDlgItemTextW(hDlg, IDC_ALLKEYS, GetLangStringW(wLangSystem, STRID_ALLKEYS));
    SetDlgItemTextW(hDlg, IDC_CLOSE, GetLangStringW(wLangSystem, STRID_CLOSE));

    SendMessage(hWndName, EM_LIMITTEXT, MAX_PATH, 0);
    SendMessage(hWndCommand, EM_LIMITTEXT, MAX_PATH, 0);
    EnableWindow(hWndAddModifyHotkey, FALSE);
    EnableWindow(hWndMoveUpHotkey, FALSE);
    EnableWindow(hWndMoveDownHotkey, FALSE);
    EnableWindow(hWndDeleteHotkey, FALSE);
    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES);

    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndHotkey, 0);

    //Columns
    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_NAME);
    lvc.cx=nColumnWidth1;
    lvc.iSubItem=LVI_LIST_NAME;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_LIST_NAME, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_COMMAND);
    lvc.cx=nColumnWidth2;
    lvc.iSubItem=LVI_LIST_COMMAND;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_LIST_COMMAND, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_HOTKEY);
    lvc.cx=nColumnWidth3;
    lvc.iSubItem=LVI_LIST_HOTKEY;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_LIST_HOTKEY, (LPARAM)&lvc);

    SendMessageW(hWndCommand, CB_ADDSTRING, 0, (LPARAM)L"Command(0)");
    SendMessageW(hWndCommand, CB_ADDSTRING, 0, (LPARAM)L"Exec(`Program.exe`)");
    SendMessageW(hWndCommand, CB_ADDSTRING, 0, (LPARAM)L"Call(\"Plugin::Function\")");
    SendMessageW(hWndCommand, CB_ADDSTRING, 0, (LPARAM)L"+Call(\"Plugin::Function\")");
    SendMessageW(hWndCommand, CB_ADDSTRING, 0, (LPARAM)L"Font(\"FaceName\", 0, 10)");
    SendMessageW(hWndCommand, CB_ADDSTRING, 0, (LPARAM)L"Recode(0, 0)");
    SendMessageW(hWndCommand, CB_ADDSTRING, 0, (LPARAM)L"Insert(\"text\")");

    FillHotkeyListW(hWndList, &hHotkeysStack);
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_LIST)
    {
      HOTKEYITEMW *lpElement;

      if (((NMLISTVIEW *)lParam)->hdr.code == LVN_ITEMCHANGED)
      {
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_SELECTED)
        {
          nSelItem=((NMLISTVIEW *)lParam)->iItem;

          if (lpElement=StackGetItemByIndexW(&hHotkeysStack, nSelItem + 1))
          {
            if (nSelItem == 0)
              EnableWindow(hWndMoveUpHotkey, FALSE);
            else
              EnableWindow(hWndMoveUpHotkey, TRUE);
            if (SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0) == nSelItem + 1)
              EnableWindow(hWndMoveDownHotkey, FALSE);
            else
              EnableWindow(hWndMoveDownHotkey, TRUE);
            EnableWindow(hWndDeleteHotkey, TRUE);

            SetWindowTextW(hWndName, lpElement->wszHotkeyName);
            SetWindowTextW(hWndCommand, lpElement->wszHotkeyCommand);
            SendMessage(hWndHotkey, HKM_SETHOTKEY, lpElement->dwHotkey, 0);
            PostMessage(hDlg, AKDLL_UPDATEBUTTON, 0, 0);
          }
        }
        if (((NMLISTVIEW *)lParam)->uOldState & LVIS_SELECTED)
        {
          nSelItem=-1;

          EnableWindow(hWndMoveUpHotkey, FALSE);
          EnableWindow(hWndMoveDownHotkey, FALSE);
          EnableWindow(hWndDeleteHotkey, FALSE);
          PostMessage(hDlg, AKDLL_UPDATEBUTTON, 0, 0);
        }
      }
      else if ((int)((NMHDR *)lParam)->code == NM_DBLCLK)
      {
        if (nSelItem >= 0)
          SendMessage(hWndList, LVM_EDITLABEL, nSelItem, 0);
      }
      else if (((NMHDR *)lParam)->code == LVN_ENDLABELEDITW)
      {
        NMLVDISPINFOW *nmlvdi=(NMLVDISPINFOW *)lParam;

        if (nmlvdi->item.pszText && nmlvdi->item.pszText[0])
        {
          if (!StackGetItemByNameW(&hHotkeysStack, nmlvdi->item.pszText))
          {
            if (lpElement=StackGetItemByIndexW(&hHotkeysStack, nSelItem + 1))
            {
              lstrcpynW(lpElement->wszHotkeyName, nmlvdi->item.pszText, MAX_PATH);
              SetWindowLongW(hDlg, DWL_MSGRESULT, 1);
            }
          }
          else PostMessage(hDlg, AKDLL_MSGHOTKEYEXISTS, 0, 0);
        }
        return TRUE;
      }
    }
  }
  else if (uMsg == AKDLL_MSGHOTKEYEXISTS)
  {
    MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_NAME_EXISTS), L"Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);
  }
  else if (uMsg == AKDLL_UPDATEBUTTON)
  {
    wchar_t wszHotkeyName[MAX_PATH];
    wchar_t wszHotkeyCommand[MAX_PATH];
    HOTKEYITEMW *lpElement;
    HWND hWndFocus;
    WORD wHotkey;
    BOOL bEnable=FALSE;

    hWndFocus=GetFocus();
    SetDlgItemTextW(hDlg, IDC_ADDMODIFYHOTKEY, GetLangStringW(wLangSystem, STRID_ADD));

    if (GetControlTextW(hWndName, wszHotkeyName, MAX_PATH))
    {
      GetControlTextW(hWndCommand, wszHotkeyCommand, MAX_PATH);
      wHotkey=(WORD)SendMessage(hWndHotkey, HKM_GETHOTKEY, 0, 0);
      bEnable=TRUE;

      if (lpElement=StackGetItemByNameW(&hHotkeysStack, wszHotkeyName))
      {
        SetDlgItemTextW(hDlg, IDC_ADDMODIFYHOTKEY, GetLangStringW(wLangSystem, STRID_MODIFY));
        if (!lstrcmpW(wszHotkeyCommand, lpElement->wszHotkeyCommand) && lpElement->dwHotkey == wHotkey)
          bEnable=FALSE;
      }
    }

    if (!bEnable)
    {
      if (hWndFocus == hWndAddModifyHotkey)
        SetFocus(hWndList);
    }
    EnableWindow(hWndAddModifyHotkey, bEnable);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_NAME ||
        LOWORD(wParam) == IDC_COMMAND ||
        LOWORD(wParam) == IDC_HOTKEY)
    {
      PostMessage(hDlg, AKDLL_UPDATEBUTTON, 0, 0);
    }
    else if (LOWORD(wParam) == IDC_ADDMODIFYHOTKEY)
    {
      PLUGINFUNCTIONW *pfElement=NULL;
      HOTKEYITEMW *hiNewElement=NULL;
      HOTKEYITEMW *hiOldElement=NULL;
      HOTKEYITEMW *hiElement=NULL;
      wchar_t wszHotkeyName[MAX_PATH];
      wchar_t wszHotkeyCommand[MAX_PATH];
      wchar_t *wpText=wszHotkeyCommand;
      wchar_t *wpHotkeyCommand;
      DWORD dwAction;
      WORD wHotkey;
      int nNewIndex;
      int nOldIndex=-1;
      int nPlus;
      int a;
      int b;
      BOOL bAddHotkey=TRUE;

      GetControlTextW(hWndName, wszHotkeyName, MAX_PATH);
      GetControlTextW(hWndCommand, wszHotkeyCommand, MAX_PATH);
      wHotkey=(WORD)SendMessage(hWndHotkey, HKM_GETHOTKEY, 0, 0);
      if (hiOldElement=StackGetItemByNameW(&hHotkeysStack, wszHotkeyName))
      {
        bAddHotkey=FALSE;
        nOldIndex=StackGetItemIndex(&hHotkeysStack, hiOldElement) - 1;
      }

      //Remove quotes
      for (a=0, b=0; wszHotkeyName[b];)
      {
        if (wszHotkeyName[b] == '\"')
          ++b;
        else
          wszHotkeyName[a++]=wszHotkeyName[b++];
      }
      wszHotkeyName[a]='\0';

      if (wHotkey)
      {
        //Is hotkey already assigned via AKD_DLLADD
        if (pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, wHotkey))
        {
          wsprintfW(wszBuffer, GetLangStringW(wLangSystem, STRID_HOTKEY_EXISTS), pfElement->szFunction);
          MessageBoxW(hDlg, wszBuffer, L"Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);
          SetFocus(hWndHotkey);
          return FALSE;
        }

        //Is hotkey already assigned via HotKeys plugin
        if (hiElement=StackGetItemByHotkeyW(&hHotkeysStack, wHotkey))
        {
          if (bAddHotkey || hiOldElement != hiElement)
          {
            wsprintfW(wszBuffer, GetLangStringW(wLangSystem, STRID_HOTKEY_EXISTS), hiElement->wszHotkeyName);
            MessageBoxW(hDlg, wszBuffer, L"Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);
            SetFocus(hWndHotkey);
            return FALSE;
          }
        }
      }

      //AutoLoad
      while (*wpText == ' ') ++wpText;
      wpHotkeyCommand=wpText;

      if (*wpText == '+')
      {
        ++wpText;
        nPlus=1;
      }
      else nPlus=0;

      //Command
      dwAction=0;
      GetActionNameW(wpText, wszBuffer, BUFFER_SIZE, &wpText);

      if (!lstrcmpiW(wszBuffer, L"Command"))
        dwAction=ACT_COMMAND;
      else if (!lstrcmpiW(wszBuffer, L"Call"))
        dwAction=ACT_CALL;
      else if (!lstrcmpiW(wszBuffer, L"Exec"))
        dwAction=ACT_EXEC;
      else if (!lstrcmpiW(wszBuffer, L"Font"))
        dwAction=ACT_FONT;
      else if (!lstrcmpiW(wszBuffer, L"Recode"))
        dwAction=ACT_RECODE;
      else if (!lstrcmpiW(wszBuffer, L"Insert"))
        dwAction=ACT_INSERT;

      if (dwAction)
      {
        if (bAddHotkey)
        {
          if (hiNewElement=StackInsertHotkeyW(&hHotkeysStack))
          {
            lstrcpynW(hiNewElement->wszHotkeyName, wszHotkeyName, MAX_PATH);
            hiNewElement->wszHotkeyCommand[0]='\0';
            hiNewElement->dwHotkey=wHotkey;
            hiNewElement->bSetOnStart=nPlus;
            hiNewElement->dwAction=dwAction;
            GetActionParametersW(&hiNewElement->hParamStack, wpText, &wpText);

            //Command string
            lstrcpynW(hiNewElement->wszHotkeyCommand, wpHotkeyCommand, min(wpText - wpHotkeyCommand + 1, MAX_PATH));
          }
        }
        else
        {
          hiNewElement=hiOldElement;

          if (hiNewElement)
          {
            lstrcpynW(hiNewElement->wszHotkeyName, wszHotkeyName, MAX_PATH);
            hiNewElement->wszHotkeyCommand[0]='\0';
            hiNewElement->dwHotkey=wHotkey;
            hiNewElement->bSetOnStart=nPlus;
            hiNewElement->dwAction=dwAction;
            FreeActionParameters(&hiNewElement->hParamStack);
            GetActionParametersW(&hiNewElement->hParamStack, wpText, &wpText);

            //Command string
            lstrcpynW(hiNewElement->wszHotkeyCommand, wpHotkeyCommand, min(wpText - wpHotkeyCommand + 1, MAX_PATH));
          }
        }
      }
      else MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_SYNTAX_ERROR), L"Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);

      if ((nNewIndex=SetHotkeyListItemW(hWndList, nOldIndex, hiNewElement, bAddHotkey)) >= 0)
      {
        SelHotkeyListItem(hWndList, nNewIndex);
        bListChanged=TRUE;
      }
      PostMessage(hDlg, AKDLL_UPDATEBUTTON, 0, 0);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_MOVEUPHOTKEY ||
             LOWORD(wParam) == IDC_MOVEDOWNHOTKEY)
    {
      HOTKEYITEMW *lpElement;
      int nOldIndex=nSelItem;
      int nNewIndex;

      if (nSelItem != -1)
      {
        if (lpElement=StackGetItemByIndexW(&hHotkeysStack, nOldIndex + 1))
        {
          //Delete
          SendMessage(hWndList, LVM_DELETEITEM, nOldIndex, 0);

          //Add
          if (LOWORD(wParam) == IDC_MOVEUPHOTKEY)
            nNewIndex=max(nOldIndex - 1, 0);
          else
            nNewIndex=nOldIndex + 1;
          nNewIndex=SetHotkeyListItemW(hWndList, nNewIndex, lpElement, TRUE);

          //Move item in stack
          StackMoveItem(&hHotkeysStack, nOldIndex + 1, nNewIndex + 1);

          //Select
          SetFocus(hWndList);
          SelHotkeyListItem(hWndList, nNewIndex);

          bListChanged=TRUE;
        }
      }
    }
    else if (LOWORD(wParam) == IDC_DELETEHOTKEY)
    {
      int nIndex=nSelItem;

      if (nSelItem != -1)
      {
        //Delete
        StackDeleteItemByIndexW(&hHotkeysStack, nIndex + 1);
        SendMessage(hWndList, LVM_DELETEITEM, nIndex, 0);
        nIndex=min(nIndex, SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0) - 1);

        //Select
        SetFocus(hWndList);
        SelHotkeyListItem(hWndList, nIndex);
        bListChanged=TRUE;
      }
    }
    else if (LOWORD(wParam) == IDC_ALLKEYS)
    {
      DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_ALLKEYS), hDlg, (DLGPROC)AllKeysDlgProcW);
    }
    else if (LOWORD(wParam) == IDOK)
    {
      if (GetFocus() == hWndList)
        SendMessage(hWndList, LVM_EDITLABEL, nSelItem, 0);
    }
    else if (LOWORD(wParam) == IDC_CLOSE ||
             LOWORD(wParam) == IDCANCEL)
    {
      int nWidth;

      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_LIST_NAME, 0);
      if (nWidth != nColumnWidth1)
      {
        nColumnWidth1=nWidth;
        bMainDialogRectSave=TRUE;
      }
      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_LIST_COMMAND, 0);
      if (nWidth != nColumnWidth2)
      {
        nColumnWidth2=nWidth;
        bMainDialogRectSave=TRUE;
      }
      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_LIST_HOTKEY, 0);
      if (nWidth != nColumnWidth3)
      {
        nColumnWidth3=nWidth;
        bMainDialogRectSave=TRUE;
      }

      if (bListChanged)
      {
        HOTKEYITEMW *lpElement;
        DWORD dwSize;

        for (lpElement=(HOTKEYITEMW *)hHotkeysStack.first, dwSize=0; lpElement; lpElement=lpElement->next)
        {
          dwSize+=wsprintfW(wszBuffer, L"\"%s\" %s Hotkey(%d)\r", lpElement->wszHotkeyName, lpElement->wszHotkeyCommand, lpElement->dwHotkey);
        }
        if (wszHotkeyText=(wchar_t *)GlobalAlloc(GPTR, (dwSize + 1) * sizeof(wchar_t)))
        {
          for (lpElement=(HOTKEYITEMW *)hHotkeysStack.first, dwSize=0; lpElement; lpElement=lpElement->next)
          {
            dwSize+=wsprintfW(wszHotkeyText + dwSize, L"\"%s\" %s Hotkey(%d)\r", lpElement->wszHotkeyName, lpElement->wszHotkeyCommand, lpElement->dwHotkey);
          }
          SaveOptionsW(OF_HOTKEYTEXT);
          GlobalFree((HGLOBAL)wszHotkeyText);
        }
      }
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

BOOL CALLBACK AllKeysDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HSTACK hTempStack={0};
  static HWND hWndList;
  static DIALOGRESIZE drs[]={{&hWndList, DRS_SIZE|DRS_X|DRS_Y},
                             {0, 0}};
  static CMDLIST cmdlist[]={{4101,  "IDM_FILE_NEW"},
                            {4102,  "IDM_FILE_CREATENEW"},
                            {4103,  "IDM_FILE_OPEN"},
                            {4104,  "IDM_FILE_REOPEN"},
                            {4105,  "IDM_FILE_SAVE"},
                            {4106,  "IDM_FILE_SAVEAS"},
                            {4107,  "IDM_FILE_PAGESETUP"},
                            {4108,  "IDM_FILE_PRINT"},
                            {4109,  "IDM_FILE_EXIT"},
                            {4110,  "IDM_FILE_SAVEAL"},
                            {4111,  "IDM_FILE_CLOSEAL"},
                            {4112,  "IDM_FILE_CLOSEALL_BUTACTIVE"},
                            {4113,  "IDM_FILE_SILENTPRINT"},
                            {4114,  "IDM_FILE_PRINTPREVIEW"},
                            {4151,  "IDM_EDIT_UNDO"},
                            {4152,  "IDM_EDIT_REDO"},
                            {4153,  "IDM_EDIT_CUT"},
                            {4154,  "IDM_EDIT_COPY"},
                            {4155,  "IDM_EDIT_PASTE"},
                            {4156,  "IDM_EDIT_CLEAR"},
                            {4157,  "IDM_EDIT_SELECTAL"},
                            {4158,  "IDM_EDIT_FIND"},
                            {4159,  "IDM_EDIT_FINDNEXTDOWN"},
                            {4160,  "IDM_EDIT_FINDNEXTUP"},
                            {4161,  "IDM_EDIT_REPLACE"},
                            {4162,  "IDM_EDIT_GOTOLINE"},
                            {4163,  "IDM_EDIT_INSERTCHAR"},
                            {4164,  "IDM_EDIT_INSERT_TAB"},
                            {4165,  "IDM_EDIT_INSERT_TAB_MENU"},
                            {4166,  "IDM_EDIT_DELETE_TAB"},
                            {4167,  "IDM_EDIT_DELETE_TAB_MENU"},
                            {4168,  "IDM_EDIT_INSERT_SPACE"},
                            {4169,  "IDM_EDIT_INSERT_SPACE_MENU"},
                            {4170,  "IDM_EDIT_DELETE_SPACE"},
                            {4171,  "IDM_EDIT_DELETE_SPACE_MENU"},
                            {4172,  "IDM_EDIT_DELETE_FIRST_CHAR"},
                            {4173,  "IDM_EDIT_DELETE_FIRST_CHAR_MENU"},
                            {4174,  "IDM_EDIT_DELETE_TRAILING_WHITESPACES"},
                            {4175,  "IDM_EDIT_UPPERCASE"},
                            {4176,  "IDM_EDIT_LOWERCASE"},
                            {4177,  "IDM_EDIT_SENTENCECASE"},
                            {4178,  "IDM_EDIT_TITLECASE"},
                            {4179,  "IDM_EDIT_INVERTCASE"},
                            {4180,  "IDM_EDIT_LOOPCASE"},
                            {4182,  "IDM_EDIT_RECODE"},
                            {4183,  "IDM_EDIT_INSERTDATE"},
                            {4184,  "IDM_EDIT_NEWLINE_WIN"},
                            {4185,  "IDM_EDIT_NEWLINE_UNIX"},
                            {4186,  "IDM_EDIT_NEWLINE_MAC"},
                            {4201,  "IDM_VIEW_FONT"},
                            {4202,  "IDM_VIEW_COLORS"},
                            {4204,  "IDM_VIEW_INCREASE_FONT"},
                            {4205,  "IDM_VIEW_DECREASE_FONT"},
                            {4209,  "IDM_VIEW_WORDWRAP"},
                            {4210,  "IDM_VIEW_ONTOP"},
                            {4211,  "IDM_VIEW_SHOW_STATUSBAR"},
                            {4212,  "IDM_VIEW_SPLIT_WINDOW_AL"},
                            {4213,  "IDM_VIEW_SPLIT_WINDOW_WE"},
                            {4214,  "IDM_VIEW_SPLIT_WINDOW_NS"},
                            {4215,  "IDM_VIEW_SPLIT_WINDOW_OFF"},
                            {4251,  "IDM_OPTIONS_EXEC"},
                            {4252,  "IDM_OPTIONS_SAVETIME"},
                            {4253,  "IDM_OPTIONS_WATCHFILE"},
                            {4254,  "IDM_OPTIONS_KEEPSPACE"},
                            {4255,  "IDM_OPTIONS_SINGLEOPEN_FILE"},
                            {4256,  "IDM_OPTIONS_SINGLEOPEN_PROGRAM"},
                            {4257,  "IDM_OPTIONS_SDI"},
                            {4258,  "IDM_OPTIONS_MDI"},
                            {4259,  "IDM_OPTIONS_PLUGINS"},
                            {4260,  "IDM_OPTIONS"},
                            {4261,  "IDM_OPTIONS_READONLY"},
                            {4301,  "IDM_WINDOW_TABVIEW_TOP"},
                            {4302,  "IDM_WINDOW_TABVIEW_BOTTOM"},
                            {4303,  "IDM_WINDOW_TABVIEW_NONE"},
                            {4304,  "IDM_WINDOW_TABTYPE_STANDARD"},
                            {4305,  "IDM_WINDOW_TABTYPE_BUTTONS"},
                            {4306,  "IDM_WINDOW_TABTYPE_FLATBUTTONS"},
                            {4307,  "IDM_WINDOW_TILEHORIZONTA"},
                            {4308,  "IDM_WINDOW_TILEVERTICA"},
                            {4309,  "IDM_WINDOW_CASCADE"},
                            {4351,  "IDM_ABOUT"},
                            {4401,  "IDM_NONMENU_CHANGESIZE"},
                            {4402,  "IDM_NONMENU_DELLINE"},
                            {4403,  "IDM_NONMENU_PASTEANSI"},
                            {4404,  "IDM_NONMENU_MDINEXT"},
                            {4405,  "IDM_NONMENU_MDIPREV"},
                            {4406,  "IDM_NONMENU_MDICLOSE"},
                            {4407,  "IDM_NONMENU_REOPEN_MSG"},
                            {4408,  "IDM_NONMENU_REDETECT"},
                            {4409,  "IDM_NONMENU_REOPENAS_ANSI"},
                            {4410,  "IDM_NONMENU_REOPENAS_OEM"},
                            {4411,  "IDM_NONMENU_REOPENAS_UTF16LE"},
                            {4412,  "IDM_NONMENU_REOPENAS_UTF16BE"},
                            {4413,  "IDM_NONMENU_REOPENAS_UTF8"},
                            {4414,  "IDM_NONMENU_REOPENAS_KOIR"},
                            {4415,  "IDM_NONMENU_SAVEAS_ANSI"},
                            {4416,  "IDM_NONMENU_SAVEAS_OEM"},
                            {4417,  "IDM_NONMENU_SAVEAS_UTF16LE"},
                            {4418,  "IDM_NONMENU_SAVEAS_UTF16BE"},
                            {4419,  "IDM_NONMENU_SAVEAS_UTF8"},
                            {4420,  "IDM_NONMENU_SAVEAS_KOIR"},
                            {4421,  "IDM_NONMENU_INSERTMODE"},
                            {4422,  "IDM_NONMENU_FILECLOSE"},
                            {4423,  "IDM_NONMENU_DLGNEXT"},
                            {4424,  "IDM_NONMENU_DLGPREV"},
                            {4425,  "IDM_NONMENU_AUTOINDENT"},
                            {4426,  "IDM_NONMENU_SAVEAS_UTF8_NOBOM"},
                            {4427,  "IDM_NONMENU_CLONENEXT"},
                            {4428,  "IDM_NONMENU_CLONEPREV"},
                            {4429,  "IDM_NONMENU_COLUMNPASTE"},
                            {4430,  "IDM_NONMENU_PASTEAFTER"},
                            {4431,  "IDM_NONMENU_CANTOPEN_MSG"},
                            {5001,  "IDM_RECENT_FILES"},
                            {6001,  "IDM_LANGUAGE"},
                            {7001,  "IDM_POPUP_OPENAS"},
                            {8001,  "IDM_POPUP_SAVEAS"},
                            {10019, "IDM_SELECTWINDOW"},
                            {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNA lvc;
    HOTKEYITEMA *hiItem;
    int nInsertItem=0;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndList=GetDlgItem(hDlg, IDC_LIST);

    SetWindowTextA(hDlg, GetLangStringA(wLangSystem, STRID_ALLKEYS));

    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES);

    //Columns
    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_NAME);
    lvc.cx=250;
    lvc.iSubItem=LVI_LIST_NAME;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_LIST_NAME, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_COMMAND);
    lvc.cx=210;
    lvc.iSubItem=LVI_LIST_COMMAND;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_LIST_COMMAND, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_HOTKEY);
    lvc.cx=106;
    lvc.iSubItem=LVI_LIST_HOTKEY;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_LIST_HOTKEY, (LPARAM)&lvc);

    //Accelerators
    {
      HACCEL hAccelTables[]={hGlobalAccel, hMainAccel, 0};
      ACCEL *lpAccelArray;
      int nAccelCount;
      BYTE nMod;
      int a;
      int b;
      int c;

      for (a=0; hAccelTables[a]; ++a)
      {
        //Get table size
        nAccelCount=CopyAcceleratorTable(hAccelTables[a], NULL, 0);

        if (lpAccelArray=(ACCEL *)GlobalAlloc(GPTR, nAccelCount * sizeof(ACCEL)))
        {
          //Get table array
          CopyAcceleratorTable(hAccelTables[a], lpAccelArray, nAccelCount);

          for (b=0; b < nAccelCount; ++b)
          {
            for (c=0; cmdlist[c].nCmd; ++c)
            {
              if (lpAccelArray[b].cmd == cmdlist[c].nCmd)
              {
                if (hiItem=StackInsertHotkeyA(&hTempStack))
                {
                  //hiItem->szHotkeyName
                  lstrcpynA(hiItem->szHotkeyName, cmdlist[c].pCmdName, MAX_PATH);

                  //hiItem->szHotkeyCommand
                  wsprintfA(hiItem->szHotkeyCommand, "Command(%d)", cmdlist[c].nCmd);

                  //hiItem->dwHotkey
                  nMod=0;
                  if (lpAccelArray[b].fVirt & FSHIFT)
                    nMod|=HOTKEYF_SHIFT;
                  if (lpAccelArray[b].fVirt & FCONTROL)
                    nMod|=HOTKEYF_CONTROL;
                  if (lpAccelArray[b].fVirt & FALT)
                    nMod|=HOTKEYF_ALT;
                  if (lpAccelArray[b].fVirt & FVIRTKEY)
                  {
                    if (lpAccelArray[b].key == VK_INSERT ||
                        lpAccelArray[b].key == VK_DELETE ||
                        lpAccelArray[b].key == VK_HOME ||
                        lpAccelArray[b].key == VK_END ||
                        lpAccelArray[b].key == VK_NEXT ||
                        lpAccelArray[b].key == VK_PRIOR ||
                        lpAccelArray[b].key == VK_LEFT ||
                        lpAccelArray[b].key == VK_RIGHT ||
                        lpAccelArray[b].key == VK_UP ||
                        lpAccelArray[b].key == VK_DOWN)
                    {
                      nMod|=HOTKEYF_EXT;
                    }
                  }
                  hiItem->dwHotkey=MAKEWORD(lpAccelArray[b].key, nMod);

                  SetHotkeyListItemA(hWndList, -1, hiItem, TRUE);
                }
              }
            }
          }
          SendMessage(hWndList, LVM_SORTITEMS, LVI_LIST_COMMAND, (LPARAM)CompareFuncA);

          GlobalFree((HGLOBAL)lpAccelArray);
        }
      }
    }

    //All plugins
    {
      PLUGINFUNCTIONA *pfElement;

      for (pfElement=(PLUGINFUNCTIONA *)hPluginsStack->first; pfElement; pfElement=pfElement->next)
      {
        if (pfElement->wHotkey)
        {
          if (hiItem=StackInsertHotkeyA(&hTempStack))
          {
            //hiItem->szHotkeyName
            lstrcpynA(hiItem->szHotkeyName, pfElement->szFunction, MAX_PATH);

            //hiItem->szHotkeyCommand
            if (StrReplaceA(hiItem->szHotkeyName, "::", "", TRUE, NULL, NULL) <= 1)
              wsprintfA(hiItem->szHotkeyCommand, "Call(\"%s\")", hiItem->szHotkeyName);
            else
              hiItem->szHotkeyCommand[0]='\0';

            //hiItem->dwHotkey
            hiItem->dwHotkey=pfElement->wHotkey;

            SetHotkeyListItemA(hWndList, nInsertItem++, hiItem, TRUE);
          }
        }
      }
    }

    //Hotkeys plugin
    {
      HOTKEYITEMA *hiElement;

      for (hiElement=(HOTKEYITEMA *)hHotkeysStack.first; hiElement; hiElement=hiElement->next)
      {
        if (hiElement->dwHotkey)
        {
          if (hiItem=StackInsertHotkeyA(&hTempStack))
          {
            //hiItem->szHotkeyName
            wsprintfA(hiItem->szHotkeyName, "Hotkeys::%s", hiElement->szHotkeyName);

            //hiItem->szHotkeyCommand
            lstrcpynA(hiItem->szHotkeyCommand, hiElement->szHotkeyCommand, MAX_PATH);

            //hiItem->dwHotkey
            hiItem->dwHotkey=hiElement->dwHotkey;

            SetHotkeyListItemA(hWndList, nInsertItem++, hiItem, TRUE);
          }
        }
      }
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_LIST)
    {
      if (((NMHDR *)lParam)->code == LVN_COLUMNCLICK)
      {
        NMLISTVIEW *pnmlv=(NMLISTVIEW *)lParam;

        SendMessage(hWndList, LVM_SORTITEMS, pnmlv->iSubItem, (LPARAM)CompareFuncA);
      }
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDCANCEL)
    {
      StackFreeHotkeyA(&hTempStack);
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  if (DialogResizeMessages(&drs[0], &rcAllKeysInitDialog, &rcAllKeysCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

BOOL CALLBACK AllKeysDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HSTACK hTempStack={0};
  static HWND hWndList;
  static DIALOGRESIZE drs[]={{&hWndList, DRS_SIZE|DRS_X|DRS_Y},
                             {0, 0}};
  static CMDLIST cmdlist[]={{4101,  "IDM_FILE_NEW"},
                            {4102,  "IDM_FILE_CREATENEW"},
                            {4103,  "IDM_FILE_OPEN"},
                            {4104,  "IDM_FILE_REOPEN"},
                            {4105,  "IDM_FILE_SAVE"},
                            {4106,  "IDM_FILE_SAVEAS"},
                            {4107,  "IDM_FILE_PAGESETUP"},
                            {4108,  "IDM_FILE_PRINT"},
                            {4109,  "IDM_FILE_EXIT"},
                            {4110,  "IDM_FILE_SAVEALL"},
                            {4111,  "IDM_FILE_CLOSEALL"},
                            {4112,  "IDM_FILE_CLOSEALL_BUTACTIVE"},
                            {4113,  "IDM_FILE_SILENTPRINT"},
                            {4114,  "IDM_FILE_PRINTPREVIEW"},
                            {4151,  "IDM_EDIT_UNDO"},
                            {4152,  "IDM_EDIT_REDO"},
                            {4153,  "IDM_EDIT_CUT"},
                            {4154,  "IDM_EDIT_COPY"},
                            {4155,  "IDM_EDIT_PASTE"},
                            {4156,  "IDM_EDIT_CLEAR"},
                            {4157,  "IDM_EDIT_SELECTALL"},
                            {4158,  "IDM_EDIT_FIND"},
                            {4159,  "IDM_EDIT_FINDNEXTDOWN"},
                            {4160,  "IDM_EDIT_FINDNEXTUP"},
                            {4161,  "IDM_EDIT_REPLACE"},
                            {4162,  "IDM_EDIT_GOTOLINE"},
                            {4163,  "IDM_EDIT_INSERTCHAR"},
                            {4164,  "IDM_EDIT_INSERT_TAB"},
                            {4165,  "IDM_EDIT_INSERT_TAB_MENU"},
                            {4166,  "IDM_EDIT_DELETE_TAB"},
                            {4167,  "IDM_EDIT_DELETE_TAB_MENU"},
                            {4168,  "IDM_EDIT_INSERT_SPACE"},
                            {4169,  "IDM_EDIT_INSERT_SPACE_MENU"},
                            {4170,  "IDM_EDIT_DELETE_SPACE"},
                            {4171,  "IDM_EDIT_DELETE_SPACE_MENU"},
                            {4172,  "IDM_EDIT_DELETE_FIRST_CHAR"},
                            {4173,  "IDM_EDIT_DELETE_FIRST_CHAR_MENU"},
                            {4174,  "IDM_EDIT_DELETE_TRAILING_WHITESPACES"},
                            {4175,  "IDM_EDIT_UPPERCASE"},
                            {4176,  "IDM_EDIT_LOWERCASE"},
                            {4177,  "IDM_EDIT_SENTENCECASE"},
                            {4178,  "IDM_EDIT_TITLECASE"},
                            {4179,  "IDM_EDIT_INVERTCASE"},
                            {4180,  "IDM_EDIT_LOOPCASE"},
                            {4182,  "IDM_EDIT_RECODE"},
                            {4183,  "IDM_EDIT_INSERTDATE"},
                            {4184,  "IDM_EDIT_NEWLINE_WIN"},
                            {4185,  "IDM_EDIT_NEWLINE_UNIX"},
                            {4186,  "IDM_EDIT_NEWLINE_MAC"},
                            {4201,  "IDM_VIEW_FONT"},
                            {4202,  "IDM_VIEW_COLORS"},
                            {4204,  "IDM_VIEW_INCREASE_FONT"},
                            {4205,  "IDM_VIEW_DECREASE_FONT"},
                            {4209,  "IDM_VIEW_WORDWRAP"},
                            {4210,  "IDM_VIEW_ONTOP"},
                            {4211,  "IDM_VIEW_SHOW_STATUSBAR"},
                            {4212,  "IDM_VIEW_SPLIT_WINDOW_ALL"},
                            {4213,  "IDM_VIEW_SPLIT_WINDOW_WE"},
                            {4214,  "IDM_VIEW_SPLIT_WINDOW_NS"},
                            {4215,  "IDM_VIEW_SPLIT_WINDOW_OFF"},
                            {4251,  "IDM_OPTIONS_EXEC"},
                            {4252,  "IDM_OPTIONS_SAVETIME"},
                            {4253,  "IDM_OPTIONS_WATCHFILE"},
                            {4254,  "IDM_OPTIONS_KEEPSPACE"},
                            {4255,  "IDM_OPTIONS_SINGLEOPEN_FILE"},
                            {4256,  "IDM_OPTIONS_SINGLEOPEN_PROGRAM"},
                            {4257,  "IDM_OPTIONS_SDI"},
                            {4258,  "IDM_OPTIONS_MDI"},
                            {4259,  "IDM_OPTIONS_PLUGINS"},
                            {4260,  "IDM_OPTIONS"},
                            {4261,  "IDM_OPTIONS_READONLY"},
                            {4301,  "IDM_WINDOW_TABVIEW_TOP"},
                            {4302,  "IDM_WINDOW_TABVIEW_BOTTOM"},
                            {4303,  "IDM_WINDOW_TABVIEW_NONE"},
                            {4304,  "IDM_WINDOW_TABTYPE_STANDARD"},
                            {4305,  "IDM_WINDOW_TABTYPE_BUTTONS"},
                            {4306,  "IDM_WINDOW_TABTYPE_FLATBUTTONS"},
                            {4307,  "IDM_WINDOW_TILEHORIZONTAL"},
                            {4308,  "IDM_WINDOW_TILEVERTICAL"},
                            {4309,  "IDM_WINDOW_CASCADE"},
                            {4351,  "IDM_ABOUT"},
                            {4401,  "IDM_NONMENU_CHANGESIZE"},
                            {4402,  "IDM_NONMENU_DELLINE"},
                            {4403,  "IDM_NONMENU_PASTEANSI"},
                            {4404,  "IDM_NONMENU_MDINEXT"},
                            {4405,  "IDM_NONMENU_MDIPREV"},
                            {4406,  "IDM_NONMENU_MDICLOSE"},
                            {4407,  "IDM_NONMENU_REOPEN_MSG"},
                            {4408,  "IDM_NONMENU_REDETECT"},
                            {4409,  "IDM_NONMENU_REOPENAS_ANSI"},
                            {4410,  "IDM_NONMENU_REOPENAS_OEM"},
                            {4411,  "IDM_NONMENU_REOPENAS_UTF16LE"},
                            {4412,  "IDM_NONMENU_REOPENAS_UTF16BE"},
                            {4413,  "IDM_NONMENU_REOPENAS_UTF8"},
                            {4414,  "IDM_NONMENU_REOPENAS_KOIR"},
                            {4415,  "IDM_NONMENU_SAVEAS_ANSI"},
                            {4416,  "IDM_NONMENU_SAVEAS_OEM"},
                            {4417,  "IDM_NONMENU_SAVEAS_UTF16LE"},
                            {4418,  "IDM_NONMENU_SAVEAS_UTF16BE"},
                            {4419,  "IDM_NONMENU_SAVEAS_UTF8"},
                            {4420,  "IDM_NONMENU_SAVEAS_KOIR"},
                            {4421,  "IDM_NONMENU_INSERTMODE"},
                            {4422,  "IDM_NONMENU_FILECLOSE"},
                            {4423,  "IDM_NONMENU_DLGNEXT"},
                            {4424,  "IDM_NONMENU_DLGPREV"},
                            {4425,  "IDM_NONMENU_AUTOINDENT"},
                            {4426,  "IDM_NONMENU_SAVEAS_UTF8_NOBOM"},
                            {4427,  "IDM_NONMENU_CLONENEXT"},
                            {4428,  "IDM_NONMENU_CLONEPREV"},
                            {4429,  "IDM_NONMENU_COLUMNPASTE"},
                            {4430,  "IDM_NONMENU_PASTEAFTER"},
                            {4431,  "IDM_NONMENU_CANTOPEN_MSG"},
                            {5001,  "IDM_RECENT_FILES"},
                            {6001,  "IDM_LANGUAGE"},
                            {7001,  "IDM_POPUP_OPENAS"},
                            {8001,  "IDM_POPUP_SAVEAS"},
                            {10019, "IDM_SELECTWINDOW"},
                            {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNW lvc;
    HOTKEYITEMW *hiItem;
    int nInsertItem=0;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndList=GetDlgItem(hDlg, IDC_LIST);

    SetWindowTextW(hDlg, GetLangStringW(wLangSystem, STRID_ALLKEYS));

    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES);

    //Columns
    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_NAME);
    lvc.cx=250;
    lvc.iSubItem=LVI_LIST_NAME;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_LIST_NAME, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_COMMAND);
    lvc.cx=210;
    lvc.iSubItem=LVI_LIST_COMMAND;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_LIST_COMMAND, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_HOTKEY);
    lvc.cx=106;
    lvc.iSubItem=LVI_LIST_HOTKEY;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_LIST_HOTKEY, (LPARAM)&lvc);

    //Accelerators
    {
      HACCEL hAccelTables[]={hGlobalAccel, hMainAccel, 0};
      ACCEL *lpAccelArray;
      int nAccelCount;
      BYTE nMod;
      int a;
      int b;
      int c;

      for (a=0; hAccelTables[a]; ++a)
      {
        //Get table size
        nAccelCount=CopyAcceleratorTable(hAccelTables[a], NULL, 0);

        if (lpAccelArray=(ACCEL *)GlobalAlloc(GPTR, nAccelCount * sizeof(ACCEL)))
        {
          //Get table array
          CopyAcceleratorTable(hAccelTables[a], lpAccelArray, nAccelCount);

          for (b=0; b < nAccelCount; ++b)
          {
            for (c=0; cmdlist[c].nCmd; ++c)
            {
              if (lpAccelArray[b].cmd == cmdlist[c].nCmd)
              {
                if (hiItem=StackInsertHotkeyW(&hTempStack))
                {
                  //hiItem->wszHotkeyName
                  MultiByteToWideChar(CP_ACP, 0, cmdlist[c].pCmdName, -1, hiItem->wszHotkeyName, MAX_PATH);

                  //hiItem->wszHotkeyCommand
                  wsprintfW(hiItem->wszHotkeyCommand, L"Command(%d)", cmdlist[c].nCmd);

                  //hiItem->dwHotkey
                  nMod=0;
                  if (lpAccelArray[b].fVirt & FSHIFT)
                    nMod|=HOTKEYF_SHIFT;
                  if (lpAccelArray[b].fVirt & FCONTROL)
                    nMod|=HOTKEYF_CONTROL;
                  if (lpAccelArray[b].fVirt & FALT)
                    nMod|=HOTKEYF_ALT;
                  if (lpAccelArray[b].fVirt & FVIRTKEY)
                  {
                    if (lpAccelArray[b].key == VK_INSERT ||
                        lpAccelArray[b].key == VK_DELETE ||
                        lpAccelArray[b].key == VK_HOME ||
                        lpAccelArray[b].key == VK_END ||
                        lpAccelArray[b].key == VK_NEXT ||
                        lpAccelArray[b].key == VK_PRIOR ||
                        lpAccelArray[b].key == VK_LEFT ||
                        lpAccelArray[b].key == VK_RIGHT ||
                        lpAccelArray[b].key == VK_UP ||
                        lpAccelArray[b].key == VK_DOWN)
                    {
                      nMod|=HOTKEYF_EXT;
                    }
                  }
                  hiItem->dwHotkey=MAKEWORD(lpAccelArray[b].key, nMod);

                  SetHotkeyListItemW(hWndList, -1, hiItem, TRUE);
                }
              }
            }
          }
          SendMessage(hWndList, LVM_SORTITEMS, LVI_LIST_COMMAND, (LPARAM)CompareFuncW);

          GlobalFree((HGLOBAL)lpAccelArray);
        }
      }
    }

    //All plugins
    {
      PLUGINFUNCTIONW *pfElement;

      for (pfElement=(PLUGINFUNCTIONW *)hPluginsStack->first; pfElement; pfElement=pfElement->next)
      {
        if (pfElement->wHotkey)
        {
          if (hiItem=StackInsertHotkeyW(&hTempStack))
          {
            //hiItem->wszHotkeyName
            lstrcpynW(hiItem->wszHotkeyName, pfElement->szFunction, MAX_PATH);

            //hiItem->wszHotkeyCommand
            if (StrReplaceW(hiItem->wszHotkeyName, L"::", L"", TRUE, NULL, NULL) <= 1)
              wsprintfW(hiItem->wszHotkeyCommand, L"Call(\"%s\")", hiItem->wszHotkeyName);
            else
              hiItem->wszHotkeyCommand[0]='\0';

            //hiItem->dwHotkey
            hiItem->dwHotkey=pfElement->wHotkey;

            SetHotkeyListItemW(hWndList, nInsertItem++, hiItem, TRUE);
          }
        }
      }
    }

    //Hotkeys plugin
    {
      HOTKEYITEMW *hiElement;

      for (hiElement=(HOTKEYITEMW *)hHotkeysStack.first; hiElement; hiElement=hiElement->next)
      {
        if (hiElement->dwHotkey)
        {
          if (hiItem=StackInsertHotkeyW(&hTempStack))
          {
            //hiItem->wszHotkeyName
            wsprintfW(hiItem->wszHotkeyName, L"Hotkeys::%s", hiElement->wszHotkeyName);

            //hiItem->wszHotkeyCommand
            lstrcpynW(hiItem->wszHotkeyCommand, hiElement->wszHotkeyCommand, MAX_PATH);

            //hiItem->dwHotkey
            hiItem->dwHotkey=hiElement->dwHotkey;

            SetHotkeyListItemW(hWndList, nInsertItem++, hiItem, TRUE);
          }
        }
      }
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_LIST)
    {
      if (((NMHDR *)lParam)->code == LVN_COLUMNCLICK)
      {
        NMLISTVIEW *pnmlv=(NMLISTVIEW *)lParam;

        SendMessage(hWndList, LVM_SORTITEMS, pnmlv->iSubItem, (LPARAM)CompareFuncW);
      }
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDCANCEL)
    {
      StackFreeHotkeyW(&hTempStack);
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  if (DialogResizeMessages(&drs[0], &rcAllKeysInitDialog, &rcAllKeysCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

void FillHotkeyListA(HWND hWnd, HSTACK *hStack)
{
  HOTKEYITEMA *lpElement=(HOTKEYITEMA *)hStack->first;
  int i=0;

  while (lpElement)
  {
    SetHotkeyListItemA(hWnd, i++, lpElement, TRUE);

    lpElement=lpElement->next;
  }
}

void FillHotkeyListW(HWND hWnd, HSTACK *hStack)
{
  HOTKEYITEMW *lpElement=(HOTKEYITEMW *)hStack->first;
  int i=0;

  while (lpElement)
  {
    SetHotkeyListItemW(hWnd, i++, lpElement, TRUE);

    lpElement=lpElement->next;
  }
}

int SetHotkeyListItemA(HWND hWnd, int nIndex, HOTKEYITEMA *hiElement, BOOL bAdd)
{
  char szHotkey[MAX_PATH];
  LVITEMA lvi;

  if (hiElement)
  {
    if (bAdd)
    {
      lvi.mask=LVIF_PARAM;
      lvi.lParam=(LPARAM)hiElement;
      lvi.iItem=(nIndex > -1)?nIndex:0x7FFFFFFF;
      lvi.iSubItem=LVI_LIST_NAME;
      nIndex=SendMessage(hWnd, LVM_INSERTITEMA, 0, (LPARAM)&lvi);
    }

    lvi.mask=LVIF_TEXT;
    lvi.pszText=hiElement->szHotkeyName;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVI_LIST_NAME;
    SendMessage(hWnd, LVM_SETITEMA, 0, (LPARAM)&lvi);

    lvi.mask=LVIF_TEXT;
    lvi.pszText=hiElement->szHotkeyCommand;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVI_LIST_COMMAND;
    SendMessage(hWnd, LVM_SETITEMA, 0, (LPARAM)&lvi);

    GetHotkeyStringA((WORD)hiElement->dwHotkey, szHotkey);
    lvi.mask=LVIF_TEXT;
    lvi.pszText=szHotkey;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVI_LIST_HOTKEY;
    SendMessage(hWnd, LVM_SETITEMA, 0, (LPARAM)&lvi);

    return nIndex;
  }
  return -1;
}

int SetHotkeyListItemW(HWND hWnd, int nIndex, HOTKEYITEMW *hiElement, BOOL bAdd)
{
  wchar_t wszHotkey[MAX_PATH];
  LVITEMW lvi;

  if (hiElement)
  {
    if (bAdd)
    {
      lvi.mask=LVIF_PARAM;
      lvi.lParam=(LPARAM)hiElement;
      lvi.iItem=(nIndex > -1)?nIndex:0x7FFFFFFF;
      lvi.iSubItem=LVI_LIST_NAME;
      nIndex=SendMessage(hWnd, LVM_INSERTITEMW, 0, (LPARAM)&lvi);
    }

    lvi.mask=LVIF_TEXT;
    lvi.pszText=hiElement->wszHotkeyName;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVI_LIST_NAME;
    SendMessage(hWnd, LVM_SETITEMW, 0, (LPARAM)&lvi);

    lvi.mask=LVIF_TEXT;
    lvi.pszText=hiElement->wszHotkeyCommand;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVI_LIST_COMMAND;
    SendMessage(hWnd, LVM_SETITEMW, 0, (LPARAM)&lvi);

    GetHotkeyStringW((WORD)hiElement->dwHotkey, wszHotkey);
    lvi.mask=LVIF_TEXT;
    lvi.pszText=wszHotkey;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVI_LIST_HOTKEY;
    SendMessage(hWnd, LVM_SETITEMW, 0, (LPARAM)&lvi);

    return nIndex;
  }
  return -1;
}

void SelHotkeyListItem(HWND hWnd, int nIndex)
{
  LVITEMA lvi;

  lvi.stateMask=LVIS_SELECTED|LVIS_FOCUSED;
  lvi.state=LVIS_SELECTED|LVIS_FOCUSED;
  SendMessage(hWnd, LVM_SETITEMSTATE, nIndex, (LPARAM)&lvi);
}

int CALLBACK CompareFuncA(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
  HOTKEYITEMA *hiElement1=(HOTKEYITEMA *)lParam1;
  HOTKEYITEMA *hiElement2=(HOTKEYITEMA *)lParam2;

  if (lParamSort == LVI_LIST_NAME)
  {
    return lstrcmpiA(hiElement1->szHotkeyName, hiElement2->szHotkeyName);
  }
  else if (lParamSort == LVI_LIST_COMMAND)
  {
    return lstrcmpiA(hiElement1->szHotkeyCommand, hiElement2->szHotkeyCommand);
  }
  else if (lParamSort == LVI_LIST_HOTKEY)
  {
    char szHotkeyName1[MAX_PATH];
    char szHotkeyName2[MAX_PATH];

    GetHotkeyStringA((WORD)hiElement1->dwHotkey, szHotkeyName1);
    GetHotkeyStringA((WORD)hiElement2->dwHotkey, szHotkeyName2);
    return lstrcmpiA(szHotkeyName1, szHotkeyName2);
  }
  return 0;
}

int CALLBACK CompareFuncW(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
  HOTKEYITEMW *hiElement1=(HOTKEYITEMW *)lParam1;
  HOTKEYITEMW *hiElement2=(HOTKEYITEMW *)lParam2;

  if (lParamSort == LVI_LIST_NAME)
  {
    return lstrcmpiW(hiElement1->wszHotkeyName, hiElement2->wszHotkeyName);
  }
  else if (lParamSort == LVI_LIST_COMMAND)
  {
    return lstrcmpiW(hiElement1->wszHotkeyCommand, hiElement2->wszHotkeyCommand);
  }
  else if (lParamSort == LVI_LIST_HOTKEY)
  {
    wchar_t wszHotkeyName1[MAX_PATH];
    wchar_t wszHotkeyName2[MAX_PATH];

    GetHotkeyStringW((WORD)hiElement1->dwHotkey, wszHotkeyName1);
    GetHotkeyStringW((WORD)hiElement2->dwHotkey, wszHotkeyName2);
    return lstrcmpiW(wszHotkeyName1, wszHotkeyName2);
  }
  return 0;
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == AKDN_HOTKEY)
  {
    BOOL *bProcess=(BOOL *)lParam;

    if (bOldWindows)
      *bProcess=CallHotkeyA(&hHotkeysStack, wParam);
    else
      *bProcess=CallHotkeyW(&hHotkeysStack, wParam);
  }

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

void CreateHotkeyStackA(HSTACK *hStack, char *pText)
{
  HOTKEYITEMA *lpElement=NULL;
  char szHotkeyName[MAX_PATH];
  char *pCommand;
  DWORD dwAction;
  int nPlus;
  int nMinus;

  if (pText)
  {
    while (*pText)
    {
      while(1)
      {
        while (*pText == ' ' || *pText == '\t' || *pText == '\r') ++pText;

        //Skip comment line
        if (*pText == ';' || *pText == '#')
        {
          while (*pText != '\r' && *pText != '\0') ++pText;
        }
        else break;
      }

      //Name
      NextStringA(pText, szHotkeyName, MAX_PATH, &pText, &nMinus);

      //AutoLoad
      while (*pText == ' ' || *pText == '\t' || *pText == '\r') ++pText;
      pCommand=pText;

      if (*pText == '+')
      {
        ++pText;
        nPlus=1;
      }
      else nPlus=0;

      //Command
      lpElement=NULL;
      dwAction=0;
      GetActionNameA(pText, szBuffer, BUFFER_SIZE, &pText);

      if (!lstrcmpiA(szBuffer, "Command"))
        dwAction=ACT_COMMAND;
      else if (!lstrcmpiA(szBuffer, "Call"))
        dwAction=ACT_CALL;
      else if (!lstrcmpiA(szBuffer, "Exec"))
        dwAction=ACT_EXEC;
      else if (!lstrcmpiA(szBuffer, "Font"))
        dwAction=ACT_FONT;
      else if (!lstrcmpiA(szBuffer, "Recode"))
        dwAction=ACT_RECODE;
      else if (!lstrcmpiA(szBuffer, "Insert"))
        dwAction=ACT_INSERT;

      if (dwAction)
      {
        if (lpElement=StackInsertHotkeyA(hStack))
        {
          lstrcpynA(lpElement->szHotkeyName, szHotkeyName, MAX_PATH);
          lpElement->szHotkeyCommand[0]='\0';
          lpElement->dwHotkey=0;
          lpElement->bSetOnStart=nPlus;
          lpElement->dwAction=dwAction;
          GetActionParametersA(&lpElement->hParamStack, pText, &pText);

          //Command string
          lstrcpynA(lpElement->szHotkeyCommand, pCommand, min(pText - pCommand + 1, MAX_PATH));

          //Hotkey
          GetActionNameA(pText, szBuffer, BUFFER_SIZE, &pText);

          if (!lstrcmpiA(szBuffer, "Hotkey"))
          {
            while (*pText == ' ' || *pText == '\t') ++pText;

            lpElement->dwHotkey=xatoiA(pText);
            while (*pText != ')' && *pText != '\0')
              ++pText;
            if (*pText == ')')
              ++pText;
          }
        }
        else break;
      }

      while (1)
      {
        while (*pText == ' ' || *pText == '\t' || *pText == '\r') ++pText;

        //Skip comment line
        if (*pText == ';' || *pText == '#')
        {
          while (*pText != '\r' && *pText != '\0') ++pText;
        }
        else break;
      }
    }
  }
}

void CreateHotkeyStackW(HSTACK *hStack, wchar_t *wpText)
{
  HOTKEYITEMW *lpElement=NULL;
  wchar_t wszHotkeyName[MAX_PATH];
  wchar_t *wpCommand;
  DWORD dwAction;
  int nPlus;
  int nMinus;

  if (wpText)
  {
    while (*wpText)
    {
      while(1)
      {
        while (*wpText == ' ' || *wpText == '\t' || *wpText == '\r') ++wpText;

        //Skip comment line
        if (*wpText == ';' || *wpText == '#')
        {
          while (*wpText != '\r' && *wpText != '\0') ++wpText;
        }
        else break;
      }

      //Name
      NextStringW(wpText, wszHotkeyName, MAX_PATH, &wpText, &nMinus);

      //AutoLoad
      while (*wpText == ' ' || *wpText == '\t' || *wpText == '\r') ++wpText;
      wpCommand=wpText;

      if (*wpText == '+')
      {
        ++wpText;
        nPlus=1;
      }
      else nPlus=0;

      //Command
      lpElement=NULL;
      dwAction=0;
      GetActionNameW(wpText, wszBuffer, BUFFER_SIZE, &wpText);

      if (!lstrcmpiW(wszBuffer, L"Command"))
        dwAction=ACT_COMMAND;
      else if (!lstrcmpiW(wszBuffer, L"Call"))
        dwAction=ACT_CALL;
      else if (!lstrcmpiW(wszBuffer, L"Exec"))
        dwAction=ACT_EXEC;
      else if (!lstrcmpiW(wszBuffer, L"Font"))
        dwAction=ACT_FONT;
      else if (!lstrcmpiW(wszBuffer, L"Recode"))
        dwAction=ACT_RECODE;
      else if (!lstrcmpiW(wszBuffer, L"Insert"))
        dwAction=ACT_INSERT;

      if (dwAction)
      {
        if (lpElement=StackInsertHotkeyW(hStack))
        {
          lstrcpynW(lpElement->wszHotkeyName, wszHotkeyName, MAX_PATH);
          lpElement->wszHotkeyCommand[0]='\0';
          lpElement->dwHotkey=0;
          lpElement->bSetOnStart=nPlus;
          lpElement->dwAction=dwAction;
          GetActionParametersW(&lpElement->hParamStack, wpText, &wpText);

          //Command string
          lstrcpynW(lpElement->wszHotkeyCommand, wpCommand, min(wpText - wpCommand + 1, MAX_PATH));

          //Hotkey
          GetActionNameW(wpText, wszBuffer, BUFFER_SIZE, &wpText);

          if (!lstrcmpiW(wszBuffer, L"Hotkey"))
          {
            while (*wpText == ' ' || *wpText == '\t') ++wpText;

            lpElement->dwHotkey=xatoiW(wpText);
            while (*wpText != ')' && *wpText != '\0')
              ++wpText;
            if (*wpText == ')')
              ++wpText;
          }
        }
        else break;
      }

      while (1)
      {
        while (*wpText == ' ' || *wpText == '\t' || *wpText == '\r') ++wpText;

        //Skip comment line
        if (*wpText == ';' || *wpText == '#')
        {
          while (*wpText != '\r' && *wpText != '\0') ++wpText;
        }
        else break;
      }
    }
  }
}

void CALLBACK CallPluginA(void *lpParameter)
{
  PLUGINCALL *pc=(PLUGINCALL *)lpParameter;
  PLUGINCALLSENDA pcs;
  char szMessage[MAX_PATH];
  char *pFunction;
  int nParamStructSize;
  LPARAM lParam;
  BOOL bAutoLoad;

  pFunction=(char *)*(int *)pc->lpStruct;
  nParamStructSize=*(int *)(pc->lpStruct + sizeof(int));
  if (nParamStructSize > (int)sizeof(int))
    lParam=(LPARAM)(pc->lpStruct + sizeof(int));
  else
    lParam=0;

  if (pc->bSetOnStart)
  {
    pcs.pFunction=pFunction;
    pcs.bOnStart=FALSE;
    pcs.lParam=0;
    pcs.lpbAutoLoad=&bAutoLoad;

    if (SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)&pcs) != UD_FAILED)
    {
      if (bAutoLoad)
      {
        PLUGINFUNCTIONA *pf;

        pcs.pFunction=pFunction;
        pcs.bOnStart=pc->bOnStart;
        pcs.lParam=lParam;
        pcs.lpbAutoLoad=NULL;
        SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)&pcs);

        if (pf=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)pFunction, 0))
        {
          if (pf->bRunning)
            pf->bOnStart=TRUE;
          else
            pf->bOnStart=FALSE;

          SendMessage(hMainWnd, AKD_DLLSAVE, 0, DLLS_CLEAR);
        }
      }
      else
      {
        wsprintfA(szMessage, GetLangStringA(wLangSystem, STRID_AUTOLOAD), pFunction);
        MessageBoxA(hMainWnd, szMessage, "Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);
      }
    }
  }
  else
  {
    pcs.pFunction=pFunction;
    pcs.bOnStart=pc->bOnStart;
    pcs.lParam=lParam;
    pcs.lpbAutoLoad=NULL;
    SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)&pcs);
  }
  GlobalFree((HGLOBAL)pc->lpStruct);
  GlobalFree((HGLOBAL)pc);
}

void CALLBACK CallPluginW(void *lpParameter)
{
  PLUGINCALL *pc=(PLUGINCALL *)lpParameter;
  PLUGINCALLSENDW pcs;
  wchar_t wszMessage[MAX_PATH];
  wchar_t *wpFunction;
  int nParamStructSize;
  LPARAM lParam;
  BOOL bAutoLoad;

  wpFunction=(wchar_t *)*(int *)pc->lpStruct;
  nParamStructSize=*(int *)(pc->lpStruct + sizeof(int));
  if (nParamStructSize > (int)sizeof(int))
    lParam=(LPARAM)(pc->lpStruct + sizeof(int));
  else
    lParam=0;

  if (pc->bSetOnStart)
  {
    pcs.pFunction=wpFunction;
    pcs.bOnStart=FALSE;
    pcs.lParam=0;
    pcs.lpbAutoLoad=&bAutoLoad;

    if (SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)&pcs) != UD_FAILED)
    {
      if (bAutoLoad)
      {
        PLUGINFUNCTIONW *pf;

        pcs.pFunction=wpFunction;
        pcs.bOnStart=pc->bOnStart;
        pcs.lParam=lParam;
        pcs.lpbAutoLoad=NULL;
        SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)&pcs);

        if (pf=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)wpFunction, 0))
        {
          if (pf->bRunning)
            pf->bOnStart=TRUE;
          else
            pf->bOnStart=FALSE;

          SendMessage(hMainWnd, AKD_DLLSAVE, 0, DLLS_CLEAR);
        }
      }
      else
      {
        wsprintfW(wszMessage, GetLangStringW(wLangSystem, STRID_AUTOLOAD), wpFunction);
        MessageBoxW(hMainWnd, wszMessage, L"Hotkeys plugin", MB_OK|MB_ICONEXCLAMATION);
      }
    }
  }
  else
  {
    pcs.pFunction=wpFunction;
    pcs.bOnStart=pc->bOnStart;
    pcs.lParam=lParam;
    pcs.lpbAutoLoad=NULL;
    SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)&pcs);
  }
  GlobalFree((HGLOBAL)pc->lpStruct);
  GlobalFree((HGLOBAL)pc);
}

BOOL CallHotkeyA(HSTACK *hStack, WORD wHotkey)
{
  HOTKEYITEMA *lpElement;
  EDITINFO ei;
  char szCurrentFile[MAX_PATH]="";

  if (lpElement=StackGetItemByHotkeyA(hStack, wHotkey))
  {
    if (lpElement->dwAction == ACT_COMMAND)
    {
      int nCommand=0;

      nCommand=GetParameterInt(&lpElement->hParamStack, 1);

      if (nCommand)
      {
        SendMessage(hMainWnd, WM_COMMAND, nCommand, 0);
      }
    }
    else if (lpElement->dwAction == ACT_CALL)
    {
      PLUGINCALL *pc;
      unsigned char *lpStruct=NULL;

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        lstrcpynA(szCurrentFile, (char *)ei.pFile, MAX_PATH);
      }
      SetParametersExpCharA(&lpElement->hParamStack, szCurrentFile, szExeDir);

      if (CreateParametersStruct(&lpElement->hParamStack, &lpStruct))
      {
        if (pc=(PLUGINCALL *)GlobalAlloc(GPTR, sizeof(PLUGINCALL)))
        {
          pc->lpStruct=lpStruct;
          pc->bOnStart=FALSE;
          pc->bSetOnStart=lpElement->bSetOnStart;
          PostMessage(hMainWnd, AKD_CALLPROC, (WPARAM)CallPluginA, (LPARAM)pc);
        }
      }
    }
    else if (lpElement->dwAction == ACT_EXEC)
    {
      STARTUPINFOA si={0};
      PROCESS_INFORMATION pi={0};
      char *pCmdLine=NULL;
      char *pWorkDir=NULL;

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        lstrcpynA(szCurrentFile, (char *)ei.pFile, MAX_PATH);
      }
      SetParametersExpCharA(&lpElement->hParamStack, szCurrentFile, szExeDir);
      pCmdLine=(char *)GetParameterExpChar(&lpElement->hParamStack, 1);
      pWorkDir=(char *)GetParameterExpChar(&lpElement->hParamStack, 2);

      if (pCmdLine)
      {
        si.cb=sizeof(STARTUPINFO);
        if (CreateProcessA(NULL, pCmdLine, NULL, NULL, FALSE, 0, NULL, pWorkDir, &si, &pi))
        {
          CloseHandle(pi.hProcess);
          CloseHandle(pi.hThread);
        }
      }
    }
    else if (lpElement->dwAction == ACT_FONT)
    {
      EDITINFO ei;
      char *pFaceName=NULL;
      DWORD dwFontStyle=0;
      int nPointSize=0;

      SetParametersExpCharA(&lpElement->hParamStack, szCurrentFile, szExeDir);
      pFaceName=(char *)GetParameterExpChar(&lpElement->hParamStack, 1);
      dwFontStyle=GetParameterInt(&lpElement->hParamStack, 2);
      nPointSize=GetParameterInt(&lpElement->hParamStack, 3);

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        LOGFONTA lfNew={0};
        LOGFONTA *lfCur;
        HDC hDC;

        lfCur=(LOGFONTA *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)ei.hWndEdit, 0);
        memcpy(&lfNew, lfCur, sizeof(LOGFONTA));

        if (nPointSize)
        {
          if (hDC=GetDC(ei.hWndEdit))
          {
            lfNew.lfHeight=-MulDiv(nPointSize, GetDeviceCaps(hDC, LOGPIXELSY), 72);
            ReleaseDC(ei.hWndEdit, hDC);
          }
        }
        if (dwFontStyle != FS_NONE)
        {
          lfNew.lfWeight=(dwFontStyle == FS_FONTBOLD || dwFontStyle == FS_FONTBOLDITALIC)?FW_BOLD:FW_NORMAL;
          lfNew.lfItalic=(dwFontStyle == FS_FONTITALIC || dwFontStyle == FS_FONTBOLDITALIC)?TRUE:FALSE;
        }
        if (*pFaceName != '\0')
        {
          lstrcpynA(lfNew.lfFaceName, pFaceName, LF_FACESIZE);
        }
        SendMessage(hMainWnd, AKD_SETFONT, (WPARAM)ei.hWndEdit, (LPARAM)&lfNew);
      }
    }
    else if (lpElement->dwAction == ACT_RECODE)
    {
      TEXTRECODE tr;
      EDITINFO ei;

      tr.nCodePageFrom=GetParameterInt(&lpElement->hParamStack, 1);
      tr.nCodePageTo=GetParameterInt(&lpElement->hParamStack, 2);

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        SendMessage(hMainWnd, AKD_RECODESEL, (WPARAM)ei.hWndEdit, (LPARAM)&tr);
      }
    }
    else if (lpElement->dwAction == ACT_INSERT)
    {
      EDITINFO ei;
      char *pText=NULL;

      SetParametersExpCharA(&lpElement->hParamStack, szCurrentFile, szExeDir);
      pText=(char *)GetParameterExpChar(&lpElement->hParamStack, 1);

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        SendMessage(hMainWnd, AKD_REPLACESELA, (WPARAM)ei.hWndEdit, (LPARAM)pText);
      }
    }
    return TRUE;
  }
  return FALSE;
}

BOOL CallHotkeyW(HSTACK *hStack, WORD wHotkey)
{
  HOTKEYITEMW *lpElement;
  EDITINFO ei;
  wchar_t wszCurrentFile[MAX_PATH];

  if (lpElement=StackGetItemByHotkeyW(hStack, wHotkey))
  {
    if (lpElement->dwAction == ACT_COMMAND)
    {
      int nCommand=0;

      nCommand=GetParameterInt(&lpElement->hParamStack, 1);

      if (nCommand)
      {
        SendMessage(hMainWnd, WM_COMMAND, nCommand, 0);
      }
    }
    else if (lpElement->dwAction == ACT_CALL)
    {
      PLUGINCALL *pc;
      unsigned char *lpStruct=NULL;

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        lstrcpynW(wszCurrentFile, (wchar_t *)ei.pFile, MAX_PATH);
      }
      SetParametersExpCharW(&lpElement->hParamStack, wszCurrentFile, wszExeDir);

      if (CreateParametersStruct(&lpElement->hParamStack, &lpStruct))
      {
        if (pc=(PLUGINCALL *)GlobalAlloc(GPTR, sizeof(PLUGINCALL)))
        {
          pc->lpStruct=lpStruct;
          pc->bOnStart=FALSE;
          pc->bSetOnStart=lpElement->bSetOnStart;
          PostMessage(hMainWnd, AKD_CALLPROC, (WPARAM)CallPluginW, (LPARAM)pc);
        }
      }
    }
    else if (lpElement->dwAction == ACT_EXEC)
    {
      STARTUPINFOW si={0};
      PROCESS_INFORMATION pi={0};
      wchar_t *wpCmdLine=NULL;
      wchar_t *wpWorkDir=NULL;

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        lstrcpynW(wszCurrentFile, (wchar_t *)ei.pFile, MAX_PATH);
      }
      SetParametersExpCharW(&lpElement->hParamStack, wszCurrentFile, wszExeDir);
      wpCmdLine=(wchar_t *)GetParameterExpChar(&lpElement->hParamStack, 1);
      wpWorkDir=(wchar_t *)GetParameterExpChar(&lpElement->hParamStack, 2);

      if (wpCmdLine)
      {
        si.cb=sizeof(STARTUPINFOW);
        if (CreateProcessW(NULL, wpCmdLine, NULL, NULL, FALSE, 0, NULL, wpWorkDir, &si, &pi))
        {
          CloseHandle(pi.hProcess);
          CloseHandle(pi.hThread);
        }
      }
    }
    else if (lpElement->dwAction == ACT_FONT)
    {
      EDITINFO ei;
      wchar_t *wpFaceName=NULL;
      DWORD dwFontStyle=0;
      int nPointSize=0;

      SetParametersExpCharW(&lpElement->hParamStack, wszCurrentFile, wszExeDir);
      wpFaceName=(wchar_t *)GetParameterExpChar(&lpElement->hParamStack, 1);
      dwFontStyle=GetParameterInt(&lpElement->hParamStack, 2);
      nPointSize=GetParameterInt(&lpElement->hParamStack, 3);

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        LOGFONTW lfNew={0};
        LOGFONTW *lfCur;
        HDC hDC;

        lfCur=(LOGFONTW *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)ei.hWndEdit, 0);
        memcpy(&lfNew, lfCur, sizeof(LOGFONTW));

        if (nPointSize)
        {
          if (hDC=GetDC(ei.hWndEdit))
          {
            lfNew.lfHeight=-MulDiv(nPointSize, GetDeviceCaps(hDC, LOGPIXELSY), 72);
            ReleaseDC(ei.hWndEdit, hDC);
          }
        }
        if (dwFontStyle != FS_NONE)
        {
          lfNew.lfWeight=(dwFontStyle == FS_FONTBOLD || dwFontStyle == FS_FONTBOLDITALIC)?FW_BOLD:FW_NORMAL;
          lfNew.lfItalic=(dwFontStyle == FS_FONTITALIC || dwFontStyle == FS_FONTBOLDITALIC)?TRUE:FALSE;
        }
        if (*wpFaceName != '\0')
        {
          lstrcpynW(lfNew.lfFaceName, wpFaceName, LF_FACESIZE);
        }
        SendMessage(hMainWnd, AKD_SETFONT, (WPARAM)ei.hWndEdit, (LPARAM)&lfNew);
      }
    }
    else if (lpElement->dwAction == ACT_RECODE)
    {
      TEXTRECODE tr;
      EDITINFO ei;

      tr.nCodePageFrom=GetParameterInt(&lpElement->hParamStack, 1);
      tr.nCodePageTo=GetParameterInt(&lpElement->hParamStack, 2);

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        SendMessage(hMainWnd, AKD_RECODESEL, (WPARAM)ei.hWndEdit, (LPARAM)&tr);
      }
    }
    else if (lpElement->dwAction == ACT_INSERT)
    {
      EDITINFO ei;
      wchar_t *wpText=NULL;

      SetParametersExpCharW(&lpElement->hParamStack, wszCurrentFile, wszExeDir);
      wpText=(wchar_t *)GetParameterExpChar(&lpElement->hParamStack, 1);

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        SendMessage(hMainWnd, AKD_REPLACESELW, (WPARAM)ei.hWndEdit, (LPARAM)wpText);
      }
    }
    return TRUE;
  }
  return FALSE;
}

HOTKEYITEMA* StackInsertHotkeyA(HSTACK *hStack)
{
  HOTKEYITEMA *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HOTKEYITEMA));
  return lpElement;
}

HOTKEYITEMW* StackInsertHotkeyW(HSTACK *hStack)
{
  HOTKEYITEMW *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HOTKEYITEMW));
  return lpElement;
}

int StackGetItemIndex(HSTACK *hStack, void *lpElement)
{
  return StackGetIndex((stack *)hStack->first, (stack *)hStack->last, (stack *)lpElement, TRUE);
}

HOTKEYITEMA* StackGetItemByIndexA(HSTACK *hStack, int nIndex)
{
  HOTKEYITEMA *lpElement;

  if (!StackGetElement((stack *)hStack->first, (stack *)hStack->last, (stack **)&lpElement, nIndex))
    return lpElement;
  return NULL;
}

HOTKEYITEMW* StackGetItemByIndexW(HSTACK *hStack, int nIndex)
{
  HOTKEYITEMW *lpElement;

  if (!StackGetElement((stack *)hStack->first, (stack *)hStack->last, (stack **)&lpElement, nIndex))
    return lpElement;
  return NULL;
}

HOTKEYITEMA* StackGetItemByHotkeyA(HSTACK *hStack, DWORD dwHotkey)
{
  HOTKEYITEMA *lpElement=(HOTKEYITEMA *)hStack->first;

  while (lpElement)
  {
    if (lpElement->dwHotkey == dwHotkey)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

HOTKEYITEMW* StackGetItemByHotkeyW(HSTACK *hStack, DWORD dwHotkey)
{
  HOTKEYITEMW *lpElement=(HOTKEYITEMW *)hStack->first;

  while (lpElement)
  {
    if (lpElement->dwHotkey == dwHotkey)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

HOTKEYITEMA* StackGetItemByNameA(HSTACK *hStack, char *pName)
{
  HOTKEYITEMA *lpElement=(HOTKEYITEMA *)hStack->first;

  while (lpElement)
  {
    if (!lstrcmpA(lpElement->szHotkeyName,  pName))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

HOTKEYITEMW* StackGetItemByNameW(HSTACK *hStack, wchar_t *wpName)
{
  HOTKEYITEMW *lpElement=(HOTKEYITEMW *)hStack->first;

  while (lpElement)
  {
    if (!lstrcmpW(lpElement->wszHotkeyName,  wpName))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackMoveItem(HSTACK *hStack, int nOldIndex, int nNewIndex)
{
  void *lpElement=NULL;

  if (!StackGetElement((stack *)hStack->first, (stack *)hStack->last, (stack **)&lpElement, nOldIndex))
    StackMoveIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement, nNewIndex);
}

void StackDeleteItemByIndexA(HSTACK *hStack, int nIndex)
{
  HOTKEYITEMA *lpElement=NULL;

  if (!StackGetElement((stack *)hStack->first, (stack *)hStack->last, (stack **)&lpElement, nIndex))
  {
    FreeActionParameters(&lpElement->hParamStack);
    StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement);
  }
}

void StackDeleteItemByIndexW(HSTACK *hStack, int nIndex)
{
  HOTKEYITEMW *lpElement=NULL;

  if (!StackGetElement((stack *)hStack->first, (stack *)hStack->last, (stack **)&lpElement, nIndex))
  {
    FreeActionParameters(&lpElement->hParamStack);
    StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement);
  }
}

void StackFreeHotkeyA(HSTACK *hStack)
{
  HOTKEYITEMA *lpElement=(HOTKEYITEMA *)hStack->first;

  while (lpElement)
  {
    FreeActionParameters(&lpElement->hParamStack);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void StackFreeHotkeyW(HSTACK *hStack)
{
  HOTKEYITEMW *lpElement=(HOTKEYITEMW *)hStack->first;

  while (lpElement)
  {
    FreeActionParameters(&lpElement->hParamStack);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void GetActionParametersA(HSTACK *hParamStack, char *pText, char **ppText)
{
  PARAMETER *lpParameter;
  char *pParamBegin=pText;
  char *pParamEnd;
  char *pString;
  char chStopChar;
  int nStringLen;

  GetActionParameter:
  while (*pParamBegin == ' ' || *pParamBegin == '\t') ++pParamBegin;

  if (*pParamBegin == '\"' || *pParamBegin == '\'' || *pParamBegin == '`')
  {
    //String
    chStopChar=*pParamBegin++;
    nStringLen=0;

    for (pParamEnd=pParamBegin; *pParamEnd != chStopChar && *pParamEnd != '\0'; ++pParamEnd)
    {
      ++nStringLen;
    }

    if (!StackInsertIndex((stack **)&hParamStack->first, (stack **)&hParamStack->last, (stack **)&lpParameter, -1, sizeof(PARAMETER)))
    {
      if (lpParameter->lParam=(LPARAM)GlobalAlloc(GPTR, nStringLen + 1))
      {
        lpParameter->dwType=PARAM_CHAR;
        pString=(char *)lpParameter->lParam;

        for (pParamEnd=pParamBegin; *pParamEnd != chStopChar && *pParamEnd != '\0'; ++pParamEnd)
        {
          *pString++=*pParamEnd;
        }
      }
    }
  }
  else
  {
    //Number
    for (pParamEnd=pParamBegin; *pParamEnd != ',' && *pParamEnd != ')' && *pParamEnd != '\0'; ++pParamEnd);

    if (!StackInsertIndex((stack **)&hParamStack->first, (stack **)&hParamStack->last, (stack **)&lpParameter, -1, sizeof(PARAMETER)))
    {
      lpParameter->dwType=PARAM_INT;
      lpParameter->lParam=xatoiA(pParamBegin);
    }
  }

  while (*pParamEnd != ',' && *pParamEnd != ')' && *pParamEnd != '\0')
    ++pParamEnd;
  if (*pParamEnd == ',')
  {
    pParamBegin=++pParamEnd;
    goto GetActionParameter;
  }
  if (*pParamEnd == ')')
    ++pParamEnd;
  *ppText=pParamEnd;
}

void GetActionParametersW(HSTACK *hParamStack, wchar_t *wpText, wchar_t **wppText)
{
  PARAMETER *lpParameter;
  wchar_t *wpParamBegin=wpText;
  wchar_t *wpParamEnd;
  wchar_t *wpString;
  wchar_t wchStopChar;
  int nStringLen;

  GetActionParameter:
  while (*wpParamBegin == ' ' || *wpParamBegin == '\t') ++wpParamBegin;

  if (*wpParamBegin == '\"' || *wpParamBegin == '\'' || *wpParamBegin == '`')
  {
    //String
    wchStopChar=*wpParamBegin++;
    nStringLen=0;

    for (wpParamEnd=wpParamBegin; *wpParamEnd != wchStopChar && *wpParamEnd != '\0'; ++wpParamEnd)
    {
      ++nStringLen;
    }

    if (!StackInsertIndex((stack **)&hParamStack->first, (stack **)&hParamStack->last, (stack **)&lpParameter, -1, sizeof(PARAMETER)))
    {
      if (lpParameter->lParam=(LPARAM)GlobalAlloc(GPTR, (nStringLen + 1) * sizeof(wchar_t)))
      {
        lpParameter->dwType=PARAM_CHAR;
        wpString=(wchar_t *)lpParameter->lParam;

        for (wpParamEnd=wpParamBegin; *wpParamEnd != wchStopChar && *wpParamEnd != '\0'; ++wpParamEnd)
        {
          *wpString++=*wpParamEnd;
        }
      }
    }
  }
  else
  {
    //Number
    for (wpParamEnd=wpParamBegin; *wpParamEnd != ',' && *wpParamEnd != ')' && *wpParamEnd != '\0'; ++wpParamEnd);

    if (!StackInsertIndex((stack **)&hParamStack->first, (stack **)&hParamStack->last, (stack **)&lpParameter, -1, sizeof(PARAMETER)))
    {
      lpParameter->dwType=PARAM_INT;
      lpParameter->lParam=xatoiW(wpParamBegin);
    }
  }

  while (*wpParamEnd != ',' && *wpParamEnd != ')' && *wpParamEnd != '\0')
    ++wpParamEnd;
  if (*wpParamEnd == ',')
  {
    wpParamBegin=++wpParamEnd;
    goto GetActionParameter;
  }
  if (*wpParamEnd == ')')
    ++wpParamEnd;
  *wppText=wpParamEnd;
}

void SetParametersExpCharA(HSTACK *hParamStack, char *pFile, char *pExeDir)
{
  //%f -file, %d -file directory, %a -AkelPad directory, %% -%
  PARAMETER *lpParameter=(PARAMETER *)hParamStack->first;
  char szFileDir[MAX_PATH];
  char *szString;
  char *pCount;
  char *pExpandStr=NULL;
  char chTmp;
  int nFileLen;
  int nFileDirLen;
  int nExeDirLen;
  int nExpandStrLen;
  int nEnvironmentLen;

  nFileLen=lstrlenA(pFile);
  nFileDirLen=GetFileDirA(pFile, szFileDir, MAX_PATH);
  nExeDirLen=lstrlenA(pExeDir);
  nExpandStrLen=0;

  while (lpParameter)
  {
    if (lpParameter->dwType == PARAM_CHAR)
    {
      nEnvironmentLen=ExpandEnvironmentStringsA((char *)lpParameter->lParam, &chTmp, 0);

      if (szString=(char *)GlobalAlloc(GPTR, nEnvironmentLen))
      {
        //Expand environment strings
        ExpandEnvironmentStringsA((char *)lpParameter->lParam, szString, nEnvironmentLen);

        //Calculate expanded size
        for (pCount=szString; *pCount; ++pCount)
        {
          if (*pCount == '%')
          {
            if (*(pCount + 1) == '%')
            {
              ++pCount;
              ++nExpandStrLen;
              continue;
            }
            if (*(pCount + 1) == 'f' || *(pCount + 1) == 'F')
            {
              ++pCount;
              nExpandStrLen+=nFileLen;
              continue;
            }
            if (*(pCount + 1) == 'd' || *(pCount + 1) == 'D')
            {
              ++pCount;
              nExpandStrLen+=nFileDirLen;
              continue;
            }
            if (*(pCount + 1) == 'a' || *(pCount + 1) == 'A')
            {
              ++pCount;
              nExpandStrLen+=nExeDirLen;
              continue;
            }
          }
          else ++nExpandStrLen;
        }

        //Expand string
        if (lpParameter->pExpanded) GlobalFree((HGLOBAL)lpParameter->pExpanded);

        if (lpParameter->pExpanded=(unsigned char *)GlobalAlloc(GPTR, nExpandStrLen + 1))
        {
          pExpandStr=(char *)lpParameter->pExpanded;

          for (pCount=szString; *pCount; ++pCount)
          {
            if (*pCount == '%')
            {
              if (*(pCount + 1) == '%')
              {
                ++pCount;
                *pExpandStr++='%';
                continue;
              }
              if (*(pCount + 1) == 'f' || *(pCount + 1) == 'F')
              {
                ++pCount;
                lstrcpyA(pExpandStr, pFile);
                pExpandStr+=nFileLen;
                continue;
              }
              if (*(pCount + 1) == 'd' || *(pCount + 1) == 'D')
              {
                ++pCount;
                lstrcpyA(pExpandStr, szFileDir);
                pExpandStr+=nFileDirLen;
                continue;
              }
              if (*(pCount + 1) == 'a' || *(pCount + 1) == 'A')
              {
                ++pCount;
                lstrcpyA(pExpandStr, pExeDir);
                pExpandStr+=nExeDirLen;
                continue;
              }
            }
            else *pExpandStr++=*pCount;
          }
        }
        GlobalFree((HGLOBAL)szString);
      }
    }
    lpParameter=lpParameter->next;
  }
}

void SetParametersExpCharW(HSTACK *hParamStack, wchar_t *wpFile, wchar_t *wpExeDir)
{
  //%f -file, %d -file directory, %a -AkelPad directory, %% -%
  PARAMETER *lpParameter=(PARAMETER *)hParamStack->first;
  wchar_t wszFileDir[MAX_PATH];
  wchar_t *wszString;
  wchar_t *wpCount;
  wchar_t *wpExpandStr;
  int nFileLen;
  int nFileDirLen;
  int nExeDirLen;
  int nExpandStrLen;
  int nEnvironmentLen;

  nFileLen=lstrlenW(wpFile);
  nFileDirLen=GetFileDirW(wpFile, wszFileDir, MAX_PATH);
  nExeDirLen=lstrlenW(wpExeDir);
  nExpandStrLen=0;

  while (lpParameter)
  {
    if (lpParameter->dwType == PARAM_CHAR)
    {
      nEnvironmentLen=ExpandEnvironmentStringsW((wchar_t *)lpParameter->lParam, NULL, 0);

      if (wszString=(wchar_t *)GlobalAlloc(GPTR, nEnvironmentLen * sizeof(wchar_t)))
      {
        //Expand environment strings
        ExpandEnvironmentStringsW((wchar_t *)lpParameter->lParam, wszString, nEnvironmentLen);

        //Calculate expanded size
        for (wpCount=wszString; *wpCount; ++wpCount)
        {
          if (*wpCount == '%')
          {
            if (*(wpCount + 1) == '%')
            {
              ++wpCount;
              ++nExpandStrLen;
              continue;
            }
            if (*(wpCount + 1) == 'f' || *(wpCount + 1) == 'F')
            {
              ++wpCount;
              nExpandStrLen+=nFileLen;
              continue;
            }
            if (*(wpCount + 1) == 'd' || *(wpCount + 1) == 'D')
            {
              ++wpCount;
              nExpandStrLen+=nFileDirLen;
              continue;
            }
            if (*(wpCount + 1) == 'a' || *(wpCount + 1) == 'A')
            {
              ++wpCount;
              nExpandStrLen+=nExeDirLen;
              continue;
            }
          }
          else ++nExpandStrLen;
        }

        //Expand string
        if (lpParameter->pExpanded) GlobalFree((HGLOBAL)lpParameter->pExpanded);

        if (lpParameter->pExpanded=(unsigned char *)GlobalAlloc(GPTR, (nExpandStrLen + 1) * sizeof(wchar_t)))
        {
          wpExpandStr=(wchar_t *)lpParameter->pExpanded;

          for (wpCount=wszString; *wpCount; ++wpCount)
          {
            if (*wpCount == '%')
            {
              if (*(wpCount + 1) == '%')
              {
                ++wpCount;
                *wpExpandStr++='%';
                continue;
              }
              if (*(wpCount + 1) == 'f' || *(wpCount + 1) == 'F')
              {
                ++wpCount;
                lstrcpyW(wpExpandStr, wpFile);
                wpExpandStr+=nFileLen;
                continue;
              }
              if (*(wpCount + 1) == 'd' || *(wpCount + 1) == 'D')
              {
                ++wpCount;
                lstrcpyW(wpExpandStr, wszFileDir);
                wpExpandStr+=nFileDirLen;
                continue;
              }
              if (*(wpCount + 1) == 'a' || *(wpCount + 1) == 'A')
              {
                ++wpCount;
                lstrcpyW(wpExpandStr, wpExeDir);
                wpExpandStr+=nExeDirLen;
                continue;
              }
            }
            else *wpExpandStr++=*wpCount;
          }
        }
        GlobalFree((HGLOBAL)wszString);
      }
    }
    lpParameter=lpParameter->next;
  }
}

int CreateParametersStruct(HSTACK *hParamStack, unsigned char **lppStruct)
{
  PARAMETER *lpParameter=(PARAMETER *)hParamStack->first;
  unsigned char *lpStruct=NULL;
  int nStructSize=0;
  int nOffset=0;
  int nElements=0;
  int nIndex=0;

  if (nElements=StackSize((stack *)hParamStack->first, (stack *)hParamStack->last))
  {
    nStructSize=(nElements + 1) * sizeof(int);

    if (lpStruct=(unsigned char *)GlobalAlloc(GPTR, nStructSize))
    {
      while (lpParameter)
      {
        if (nIndex == 1)
        {
          //nIndex == 1 is the size of the call parameters structure
          *((int *)(lpStruct + nOffset))=nElements * sizeof(int);
          nOffset+=sizeof(int);
          ++nIndex;
        }

        //nIndex == 0 is function name, nIndex >= 2 is call parameters
        if (lpParameter->dwType == PARAM_CHAR)
          *((int *)(lpStruct + nOffset))=(int)lpParameter->pExpanded;
        else
          *((int *)(lpStruct + nOffset))=lpParameter->lParam;

        nOffset+=sizeof(int);
        ++nIndex;
        lpParameter=lpParameter->next;
      }
    }
  }
  *lppStruct=lpStruct;
  return nStructSize;
}

int GetParameterInt(HSTACK *hParamStack, int nIndex)
{
  PARAMETER *lpParameter;

  if (!StackGetElement((stack *)hParamStack->first, (stack *)hParamStack->last, (stack **)&lpParameter, nIndex))
  {
    if (lpParameter->dwType == PARAM_INT)
      return lpParameter->lParam;
  }
  return 0;
}

unsigned char* GetParameterChar(HSTACK *hParamStack, int nIndex)
{
  PARAMETER *lpParameter;

  if (!StackGetElement((stack *)hParamStack->first, (stack *)hParamStack->last, (stack **)&lpParameter, nIndex))
  {
    if (lpParameter->dwType == PARAM_CHAR)
      return (unsigned char *)lpParameter->lParam;
  }
  return NULL;
}

unsigned char* GetParameterExpChar(HSTACK *hParamStack, int nIndex)
{
  PARAMETER *lpParameter;

  if (!StackGetElement((stack *)hParamStack->first, (stack *)hParamStack->last, (stack **)&lpParameter, nIndex))
  {
    if (lpParameter->dwType == PARAM_CHAR)
      return lpParameter->pExpanded;
  }
  return NULL;
}

void FreeActionParameters(HSTACK *hParamStack)
{
  PARAMETER *lpParameter=(PARAMETER *)hParamStack->first;

  while (lpParameter)
  {
    if (lpParameter->dwType == PARAM_CHAR)
    {
      if (lpParameter->lParam) GlobalFree((HGLOBAL)lpParameter->lParam);
      if (lpParameter->pExpanded) GlobalFree((HGLOBAL)lpParameter->pExpanded);
    }
    lpParameter=lpParameter->next;
  }
  StackClear((stack **)&hParamStack->first, (stack **)&hParamStack->last);
}

int NextStringA(char *pText, char *szStr, int nStrLen, char **ppText, int *nMinus)
{
  int i;

  while (*pText == ' ' || *pText == '\t' || *pText == '\r') ++pText;

  if (*pText == '-')
  {
    if (nMinus) *nMinus=1;
    ++pText;
  }
  else
  {
    if (nMinus) *nMinus=0;
  }

  if (*pText == '\"')
  {
    ++pText;
    i=0;

    while (*pText != '\"' && *pText != '\0')
    {
      if (i < nStrLen) szStr[i++]=*pText;
      ++pText;
    }
  }
  else
  {
    i=0;

    while (*pText != ' ' && *pText != '\r' && *pText != '\0')
    {
      if (i < nStrLen) szStr[i++]=*pText;
      ++pText;
    }
  }
  szStr[i]='\0';
  if (*pText != '\r' && *pText != '\0') ++pText;
  if (ppText) *ppText=pText;
  return i;
}

int NextStringW(wchar_t *wpText, wchar_t *wszStr, int nStrLen, wchar_t **wppText, int *nMinus)
{
  int i;

  while (*wpText == ' ' || *wpText == '\t' || *wpText == '\r') ++wpText;

  if (*wpText == '-')
  {
    if (nMinus) *nMinus=1;
    ++wpText;
  }
  else
  {
    if (nMinus) *nMinus=0;
  }

  if (*wpText == '\"')
  {
    ++wpText;
    i=0;

    while (*wpText != '\"' && *wpText != '\0')
    {
      if (i < nStrLen) wszStr[i++]=*wpText;
      ++wpText;
    }
  }
  else
  {
    i=0;

    while (*wpText != ' ' && *wpText != '\r' && *wpText != '\0')
    {
      if (i < nStrLen) wszStr[i++]=*wpText;
      ++wpText;
    }
  }
  wszStr[i]='\0';
  if (*wpText != '\r' && *wpText != '\0') ++wpText;
  if (wppText) *wppText=wpText;
  return i;
}

int GetActionNameA(char *pText, char *szStr, int nStrLen, char **ppText)
{
  int i=0;

  while (*pText == '\"' || *pText == ' ' || *pText == '\t') ++pText;

  while (*pText != '(' && *pText != '\r' && *pText != '\0')
  {
    if (i < nStrLen) szStr[i++]=*pText;
    ++pText;
  }
  szStr[i]='\0';
  if (*pText != '\r' && *pText != '\0') ++pText;
  if (ppText) *ppText=pText;
  return i;
}

int GetActionNameW(wchar_t *wpText, wchar_t *wszStr, int nStrLen, wchar_t **wppText)
{
  int i=0;

  while (*wpText == '\"' || *wpText == ' ' || *wpText == '\t') ++wpText;

  while (*wpText != '(' && *wpText != '\r' && *wpText != '\0')
  {
    if (i < nStrLen) wszStr[i++]=*wpText;
    ++wpText;
  }
  wszStr[i]='\0';
  if (*wpText != '\r' && *wpText != '\0') ++wpText;
  if (wppText) *wppText=wpText;
  return i;
}

int GetFileDirA(char *pFile, char *szFileDir, int nFileDirLen)
{
  int i;

  if (!nFileDirLen) return 0;
  szFileDir[0]='\0';

  for (i=lstrlenA(pFile) - 1; i > 0; --i)
  {
    if (pFile[i] == '\\')
    {
      i=min(nFileDirLen, i + 1);
      lstrcpynA(szFileDir, pFile, i);
      return (i - 1);
    }
  }
  return 0;
}

int GetFileDirW(wchar_t *wpFile, wchar_t *wszFileDir, int nFileDirLen)
{
  int i;

  if (!nFileDirLen) return 0;
  wszFileDir[0]='\0';

  for (i=lstrlenW(wpFile) - 1; i > 0; --i)
  {
    if (wpFile[i] == '\\')
    {
      i=min(nFileDirLen, i + 1);
      lstrcpynW(wszFileDir, wpFile, i);
      return (i - 1);
    }
  }
  return 0;
}

void GetHotkeyStringA(WORD wHotkey, char *szString)
{
  char szKeyText[100];
  LONG lScan;

  szString[0]='\0';

  if (wHotkey)
  {
    if (HIBYTE(wHotkey) & HOTKEYF_CONTROL)
      lstrcatA(szString, "Ctrl + ");
    if (HIBYTE(wHotkey) & HOTKEYF_SHIFT)
      lstrcatA(szString, "Shift + ");
    if (HIBYTE(wHotkey) & HOTKEYF_ALT)
      lstrcatA(szString, "Alt + ");
    lScan=MapVirtualKey(LOBYTE(wHotkey), 0) << 16;
    if (HIBYTE(wHotkey) & HOTKEYF_EXT) lScan|=0x01000000L;
    GetKeyNameTextA(lScan, szKeyText, 100);
    lstrcatA(szString, szKeyText);
  }
}

void GetHotkeyStringW(WORD wHotkey, wchar_t *wszString)
{
  wchar_t wszKeyText[100];
  LONG lScan;

  wszString[0]='\0';

  if (wHotkey)
  {
    if (HIBYTE(wHotkey) & HOTKEYF_CONTROL)
      lstrcatW(wszString, L"Ctrl + ");
    if (HIBYTE(wHotkey) & HOTKEYF_SHIFT)
      lstrcatW(wszString, L"Shift + ");
    if (HIBYTE(wHotkey) & HOTKEYF_ALT)
      lstrcatW(wszString, L"Alt + ");
    lScan=MapVirtualKeyW(LOBYTE(wHotkey), 0) << 16;
    if (HIBYTE(wHotkey) & HOTKEYF_EXT) lScan|=0x01000000L;
    GetKeyNameTextW(lScan, wszKeyText, 100);
    lstrcatW(wszString, wszKeyText);
  }
}

int GetControlTextA(HWND hWnd, char *szText, int nTextMax)
{
  char szBuffer[MAX_PATH];
  int nTextLen;
  int i;

  nTextLen=GetWindowTextA(hWnd, szBuffer, MAX_PATH);

  //Remove trailing spaces
  while (nTextLen > 0 && szBuffer[nTextLen - 1] == ' ')
  {
    szText[--nTextLen]='\0';
  }

  //Count leading spaces
  for (i=0; i < nTextLen; ++i)
  {
    if (szBuffer[i] != ' ')
      break;
  }
  nTextLen-=i;
  nTextLen=min(nTextMax - 1, nTextLen);
  lstrcpynA(szText, szBuffer + i, min(nTextMax, nTextLen + 1));
  return nTextLen;
}

int GetControlTextW(HWND hWnd, wchar_t *wszText, int nTextMax)
{
  wchar_t wszBuffer[MAX_PATH];
  int nTextLen;
  int i;

  nTextLen=GetWindowTextW(hWnd, wszBuffer, MAX_PATH);

  //Remove trailing spaces
  while (nTextLen > 0 && wszBuffer[nTextLen - 1] == ' ')
  {
    wszText[--nTextLen]='\0';
  }

  //Count leading spaces
  for (i=0; i < nTextLen; ++i)
  {
    if (wszBuffer[i] != ' ')
      break;
  }
  nTextLen-=i;
  nTextLen=min(nTextMax - 1, nTextLen);
  lstrcpynW(wszText, wszBuffer + i, min(nTextMax, nTextLen + 1));
  return nTextLen;
}

BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc)
{
  if (GetWindowRect(hWnd, rc))
  {
    rc->right-=rc->left;
    rc->bottom-=rc->top;

    if (hWndOwner)
    {
      if (!ScreenToClient(hWndOwner, (POINT *)&rc->left))
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG)
  {
    RECT rcTemplate;

    GetWindowPos(hDlg, NULL, rcInit);
    rcTemplate=*rcCurrent;
    *rcCurrent=*rcInit;

    if (rcTemplate.right && rcTemplate.bottom)
    {
      rcTemplate.left=rcInit->left + (rcInit->right - rcTemplate.right) / 2;
      rcTemplate.top=rcInit->top + (rcInit->bottom - rcTemplate.bottom) / 2;
      SetWindowPos(hDlg, 0, rcTemplate.left, rcTemplate.top, rcTemplate.right, rcTemplate.bottom, SWP_NOZORDER);
    }
  }
  else if (uMsg == WM_GETMINMAXINFO)
  {
    MINMAXINFO *mmi=(MINMAXINFO *)lParam;

    mmi->ptMinTrackSize.x=rcInit->right;
    mmi->ptMinTrackSize.y=rcInit->bottom;
  }
  else if (uMsg == WM_SIZE)
  {
    if (lParam)
    {
      RECT rcTemplate;
      RECT rcControl;
      DWORD dwFlags;
      POINT pt;
      int i;

      GetWindowPos(hDlg, NULL, &rcTemplate);
      pt.x=rcTemplate.right - rcCurrent->right;
      pt.y=rcTemplate.bottom - rcCurrent->bottom;
      *rcCurrent=rcTemplate;

      for (i=0; drs[i].lpWnd; ++i)
      {
        if (*drs[i].lpWnd)
        {
          dwFlags=0;
          if (drs[i].dwType & DRS_SIZE)
            dwFlags|=SWP_NOMOVE;
          else if (drs[i].dwType & DRS_MOVE)
            dwFlags|=SWP_NOSIZE;
          GetWindowPos(*drs[i].lpWnd, hDlg, &rcControl);
          SetWindowPos(*drs[i].lpWnd, 0, rcControl.left + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.top + ((drs[i].dwType & DRS_Y)?pt.y:0), rcControl.right + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.bottom + ((drs[i].dwType & DRS_Y)?pt.y:0), dwFlags|SWP_NOZORDER);
        }
      }
      InvalidateRect(hDlg, NULL, TRUE);
      return TRUE;
    }
  }
  else if (uMsg == WM_PAINT)
  {
    PAINTSTRUCT ps;
    RECT rcGrip;
    HDC hDC;

    if (hDC=BeginPaint(hDlg, &ps))
    {
      GetClientRect(hDlg, &rcGrip);
      rcGrip.left=rcGrip.right - GetSystemMetrics(SM_CXVSCROLL);
      rcGrip.top=rcGrip.bottom - GetSystemMetrics(SM_CYVSCROLL);
      DrawFrameControl(hDC, &rcGrip, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
      EndPaint(hDlg, &ps);
    }
  }
  return FALSE;
}


//// Options

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;
  DWORD dwSize;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="HotkeyText";
    po.dwType=PO_BINARY;
    po.lpData=NULL;
    po.dwData=0;
    dwSize=SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    if (dwSize)
    {
      if (wszHotkeyText=(wchar_t *)GlobalAlloc(GPTR, dwSize))
      {
        po.pOptionName="HotkeyText";
        po.dwType=PO_BINARY;
        po.lpData=(LPBYTE)wszHotkeyText;
        po.dwData=dwSize;
        SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
      }
    }

    po.pOptionName="WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcMainCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="ColumnWidth1";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth1;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="ColumnWidth2";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth2;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="ColumnWidth3";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth3;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AllKeysRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcAllKeysCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }

  if (!wszHotkeyText)
  {
    dwSize=wcslen(TXT_DEFAULT_HOTKEYS) * sizeof(wchar_t) + 2;

    if (wszHotkeyText=(wchar_t *)GlobalAlloc(GPTR, dwSize))
    {
      memcpy((unsigned char *)wszHotkeyText, TXT_DEFAULT_HOTKEYS, dwSize);
    }
  }

  //Convert Unicode to Ansi
  if (wszHotkeyText)
  {
    dwSize=WideCharToMultiByte(CP_ACP, 0, wszHotkeyText, -1, NULL, 0, NULL, NULL);

    if (szHotkeyText=(char *)GlobalAlloc(GPTR, dwSize))
    {
      WideCharToMultiByte(CP_ACP, 0, wszHotkeyText, -1, szHotkeyText, dwSize, NULL, NULL);
    }
    GlobalFree((HGLOBAL)wszHotkeyText);
    wszHotkeyText=NULL;
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;
  DWORD dwSize;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"HotkeyText";
    po.dwType=PO_BINARY;
    po.lpData=NULL;
    po.dwData=0;
    dwSize=SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    if (dwSize)
    {
      if (wszHotkeyText=(wchar_t *)GlobalAlloc(GPTR, dwSize))
      {
        po.pOptionName=L"HotkeyText";
        po.dwType=PO_BINARY;
        po.lpData=(LPBYTE)wszHotkeyText;
        po.dwData=dwSize;
        SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
      }
    }

    po.pOptionName=L"WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcMainCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"ColumnWidth1";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth1;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"ColumnWidth2";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth2;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"ColumnWidth3";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth3;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AllKeysRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcAllKeysCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }

  if (!wszHotkeyText)
  {
    dwSize=wcslen(TXT_DEFAULT_HOTKEYS) * sizeof(wchar_t) + 2;

    if (wszHotkeyText=(wchar_t *)GlobalAlloc(GPTR, dwSize))
    {
      memcpy((unsigned char *)wszHotkeyText, TXT_DEFAULT_HOTKEYS, dwSize);
    }
  }
}

void SaveOptionsA(DWORD dwFlags)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;
  DWORD dwSize;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    if (dwFlags & OF_HOTKEYTEXT)
    {
      if (szHotkeyText)
      {
        dwSize=MultiByteToWideChar(CP_ACP, 0, szHotkeyText, -1, NULL, 0) * sizeof(wchar_t);

        if (wszHotkeyText=(wchar_t *)GlobalAlloc(GPTR, dwSize))
        {
          MultiByteToWideChar(CP_ACP, 0, szHotkeyText, -1, wszHotkeyText, dwSize / sizeof(wchar_t));

          po.pOptionName="HotkeyText";
          po.dwType=PO_BINARY;
          po.lpData=(LPBYTE)wszHotkeyText;
          po.dwData=wcslen(wszHotkeyText) * sizeof(wchar_t) + 2;
          SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

          GlobalFree((HGLOBAL)wszHotkeyText);
          wszHotkeyText=NULL;
        }
      }
    }

    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName="WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcMainCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="ColumnWidth1";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth1;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="ColumnWidth2";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth2;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="ColumnWidth3";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth3;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AllKeysRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcAllKeysCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW(DWORD dwFlags)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    if (dwFlags & OF_HOTKEYTEXT)
    {
      po.pOptionName=L"HotkeyText";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)wszHotkeyText;
      po.dwData=wcslen(wszHotkeyText) * sizeof(wchar_t) + 2;
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName=L"WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcMainCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"ColumnWidth1";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth1;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"ColumnWidth2";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth2;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"ColumnWidth3";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth3;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AllKeysRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcAllKeysCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_NAME)
      return L"\x0418\x043C\x044F";
    if (nStringID == STRID_COMMAND)
      return L"\x041A\x043E\x043C\x0430\x043D\x0434\x0430";
    if (nStringID == STRID_HOTKEY)
      return L"\x0413\x043E\x0440\x044F\x0447\x0430\x044F\x0020\x043A\x043B\x0430\x0432\x0438\x0448\x0430";
    if (nStringID == STRID_ADD)
      return L"\x0414\x043E\x0431\x0430\x0432\x0438\x0442\x044C";
    if (nStringID == STRID_MODIFY)
      return L"\x0418\x0437\x043C\x0435\x043D\x0438\x0442\x044C";
    if (nStringID == STRID_DELETE)
      return L"\x0423\x0434\x0430\x043B\x0438\x0442\x044C";
    if (nStringID == STRID_ALLKEYS)
      return L"\x0412\x0441\x0435\x0020\x043A\x043B\x0430\x0432\x0438\x0448\x0438\x002E\x002E\x002E";
    if (nStringID == STRID_MOVEUP)
      return L"\x0412\x0432\x0435\x0440\x0445";
    if (nStringID == STRID_MOVEDOWN)
      return L"\x0412\x043D\x0438\x0437";
    if (nStringID == STRID_CLOSE)
      return L"\x0417\x0430\x043A\x0440\x044B\x0442\x044C";
    if (nStringID == STRID_AUTOLOAD)
      return L"\x0022\x0025\x0073\x0022\x0020\x043D\x0435\x0020\x043F\x043E\x0434\x0434\x0435\x0440\x0436\x0438\x0432\x0430\x0435\x0442\x0020\x0430\x0432\x0442\x043E\x0437\x0430\x0433\x0440\x0443\x0437\x043A\x0443\x0020\x0028\x0022\x002B\x0022\x0020\x043F\x0435\x0440\x0435\x0434\x0020\x0022\x0043\x0061\x006C\x006C\x0022\x0029\x002E";
    if (nStringID == STRID_SYNTAX_ERROR)
      return L"\x0421\x0438\x043D\x0442\x0430\x043A\x0441\x0438\x0447\x0435\x0441\x043A\x0430\x044F\x0020\x043E\x0448\x0438\x0431\x043A\x0430\x002E";
    if (nStringID == STRID_NAME_EXISTS)
      return L"\x0418\x043C\x044F\x0020\x0443\x0436\x0435\x0020\x0441\x0443\x0449\x0435\x0441\x0442\x0432\x0443\x0435\x0442\x002E";
    if (nStringID == STRID_HOTKEY_EXISTS)
      return L"\x0413\x043E\x0440\x044F\x0447\x0430\x044F\x0020\x043A\x043B\x0430\x0432\x0438\x0448\x0430\x0020\x0443\x0436\x0435\x0020\x043D\x0430\x0437\x043D\x0430\x0447\x0435\x043D\x0430\x0020\x0434\x043B\x044F\x0020\x0022\x0025\x0073\x0022\x002E";
  }
  else
  {
    if (nStringID == STRID_NAME)
      return L"Name";
    if (nStringID == STRID_COMMAND)
      return L"Command";
    if (nStringID == STRID_HOTKEY)
      return L"Hotkey";
    if (nStringID == STRID_ADD)
      return L"Add";
    if (nStringID == STRID_MODIFY)
      return L"Modify";
    if (nStringID == STRID_DELETE)
      return L"Delete";
    if (nStringID == STRID_ALLKEYS)
      return L"All keys...";
    if (nStringID == STRID_MOVEUP)
      return L"Up";
    if (nStringID == STRID_MOVEDOWN)
      return L"Down";
    if (nStringID == STRID_CLOSE)
      return L"Close";
    if (nStringID == STRID_AUTOLOAD)
      return L"\"%s\" doesn't support autoload (\"+\" before \"Call\").";
    if (nStringID == STRID_SYNTAX_ERROR)
      return L"Syntax error.";
    if (nStringID == STRID_NAME_EXISTS)
      return L"\"%s\" name already exits.";
    if (nStringID == STRID_HOTKEY_EXISTS)
      return L"Hotkey already assigned to \"%s\".";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hInstanceEXE=pd->hInstanceEXE;
  hPluginsStack=pd->hPluginsStack;
  hMainWnd=pd->hMainWnd;
  hMdiClient=pd->hMdiClient;
  hMainMenu=pd->hMainMenu;
  hMenuRecentFiles=pd->hMenuRecentFiles;
  hMainIcon=pd->hMainIcon;
  hGlobalAccel=pd->hGlobalAccel;
  hMainAccel=pd->hMainAccel;
  bOldWindows=pd->bOldWindows;
  bMDI=pd->bMDI;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);
  hPopupEdit=GetSubMenu(pd->hPopupMenu, 0);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
  {
    lstrcpynA(szExeDir, (char *)pd->pAkelDir, MAX_PATH);
    ReadOptionsA();
    CreateHotkeyStackA(&hHotkeysStack, szHotkeyText);
    GlobalFree((HGLOBAL)szHotkeyText);
  }
  else
  {
    lstrcpynW(wszExeDir, (wchar_t *)pd->pAkelDir, MAX_PATH);
    ReadOptionsW();
    CreateHotkeyStackW(&hHotkeysStack, wszHotkeyText);
    GlobalFree((HGLOBAL)wszHotkeyText);
  }
}

void InitMain()
{
  bInitMain=TRUE;

  //SubClass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProc, (LPARAM)&NewMainProcData);
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }

  if (bOldWindows)
  {
    if (bMainDialogRectSave) SaveOptionsA(OF_WINDOWRECT);
  }
  else
  {
    if (bMainDialogRectSave) SaveOptionsW(OF_WINDOWRECT);
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain)
    {
      if (bOldWindows)
        StackFreeHotkeyA(&hHotkeysStack);
      else
        StackFreeHotkeyW(&hHotkeysStack);
      UninitMain();
    }
  }
  return TRUE;
}
